import unittest
from openmdao.utils.testing_utils import use_tempdirs

import openmdao.api as om
from openmdao.utils.general_utils import set_pyoptsparse_opt
from openmdao.utils.assert_utils import assert_near_equal
import dymos as dm
from dymos.examples.balanced_field.balanced_field_ode import BalancedFieldODEComp


def _run_balanced_field_length_problem(tx=dm.GaussLobatto, timeseries=True, sim=True, solvesegs=False):
    p = om.Problem()

    _, optimizer = set_pyoptsparse_opt('IPOPT', fallback=True)

    p.driver = om.pyOptSparseDriver()
    p.driver.declare_coloring()

    # Use IPOPT if available, with fallback to SLSQP
    p.driver.options['optimizer'] = optimizer
    p.driver.options['print_results'] = False
    if optimizer == 'IPOPT':
        p.driver.opt_settings['print_level'] = 5

    # First Phase: Brake release to V1 - both engines operable
    br_to_v1 = dm.Phase(ode_class=BalancedFieldODEComp,
                        transcription=tx(num_segments=3, solve_segments=solvesegs),
                        ode_init_kwargs={'mode': 'runway'})
    br_to_v1.set_time_options(fix_initial=True, duration_bounds=(1, 1000), duration_ref=10.0)
    br_to_v1.add_state('r', fix_initial=True, lower=0, ref=1000.0, defect_ref=1000.0)
    br_to_v1.add_state('v', fix_initial=True, lower=0, ref=100.0, defect_ref=100.0)
    br_to_v1.add_parameter('alpha', val=0.0, opt=False, units='deg')

    if timeseries:
        br_to_v1.add_timeseries_output('*')

    # Second Phase: Rejected takeoff at V1 - no engines operable
    rto = dm.Phase(ode_class=BalancedFieldODEComp,
                   transcription=dm.Radau(num_segments=3, solve_segments=solvesegs),
                   ode_init_kwargs={'mode': 'runway'})
    rto.set_time_options(fix_initial=False, duration_bounds=(1, 1000), duration_ref=1.0)
    rto.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
    rto.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
    rto.add_parameter('alpha', val=0.0, opt=False, units='deg')

    if timeseries:
        rto.add_timeseries_output('*')

    # Third Phase: V1 to Vr - single engine operable
    v1_to_vr = dm.Phase(ode_class=BalancedFieldODEComp,
                        transcription=dm.Radau(num_segments=3, solve_segments=solvesegs),
                        ode_init_kwargs={'mode': 'runway'})
    v1_to_vr.set_time_options(fix_initial=False, duration_bounds=(1, 1000), duration_ref=1.0)
    v1_to_vr.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
    v1_to_vr.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
    v1_to_vr.add_parameter('alpha', val=0.0, opt=False, units='deg')

    if timeseries:
        v1_to_vr.add_timeseries_output('*')

    # Fourth Phase: Rotate - single engine operable
    rotate = dm.Phase(ode_class=BalancedFieldODEComp,
                      transcription=dm.Radau(num_segments=3, solve_segments=solvesegs),
                      ode_init_kwargs={'mode': 'runway'})
    rotate.set_time_options(fix_initial=False, duration_bounds=(1.0, 5), duration_ref=1.0)
    rotate.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
    rotate.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
    rotate.add_polynomial_control('alpha', order=1, opt=True, units='deg', lower=0, upper=10,
                                  ref=10, val=[0, 10])

    if timeseries:
        rotate.add_timeseries_output('*')

    # Fifth Phase: Climb to target speed and altitude at end of runway.
    climb = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=5),
                     ode_init_kwargs={'mode': 'climb'})
    climb.set_time_options(fix_initial=False, duration_bounds=(1, 100), duration_ref=1.0)
    climb.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
    climb.add_state('h', fix_initial=True, lower=0, ref=1.0, defect_ref=1.0)
    climb.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
    climb.add_state('gam', fix_initial=True, lower=0, ref=0.05, defect_ref=0.05)
    climb.add_control('alpha', opt=True, units='deg', lower=-10, upper=15, ref=10)

    if timeseries:
        climb.add_timeseries_output('*')

    # Instantiate the trajectory and add phases
    traj = dm.Trajectory()
    p.model.add_subsystem('traj', traj)
    traj.add_phase('br_to_v1', br_to_v1)
    traj.add_phase('rto', rto)
    traj.add_phase('v1_to_vr', v1_to_vr)
    traj.add_phase('rotate', rotate)
    traj.add_phase('climb', climb)

    # Add parameters common to multiple phases to the trajectory
    traj.add_parameter('m', val=174200., opt=False, units='lbm',
                       desc='aircraft mass',
                       targets={'br_to_v1': ['m'], 'v1_to_vr': ['m'], 'rto': ['m'],
                                'rotate': ['m'], 'climb': ['m']})

    traj.add_parameter('T_nominal', val=27000 * 2, opt=False, units='lbf', static_target=True,
                       desc='nominal aircraft thrust',
                       targets={'br_to_v1': ['T']})

    traj.add_parameter('T_engine_out', val=27000, opt=False, units='lbf', static_target=True,
                       desc='thrust under a single engine',
                       targets={'v1_to_vr': ['T'], 'rotate': ['T'], 'climb': ['T']})

    traj.add_parameter('T_shutdown', val=0.0, opt=False, units='lbf', static_target=True,
                       desc='thrust when engines are shut down for rejected takeoff',
                       targets={'rto': ['T']})

    traj.add_parameter('mu_r_nominal', val=0.03, opt=False, units=None, static_target=True,
                       desc='nominal runway friction coefficient',
                       targets={'br_to_v1': ['mu_r'], 'v1_to_vr': ['mu_r'], 'rotate': ['mu_r']})

    traj.add_parameter('mu_r_braking', val=0.3, opt=False, units=None, static_target=True,
                       desc='runway friction coefficient under braking',
                       targets={'rto': ['mu_r']})

    traj.add_parameter('h_runway', val=0., opt=False, units='ft',
                       desc='runway altitude',
                       targets={'br_to_v1': ['h'], 'v1_to_vr': ['h'], 'rto': ['h'],
                                'rotate': ['h']})

    traj.add_parameter('rho', val=1.225, opt=False, units='kg/m**3', static_target=True,
                       desc='atmospheric density',
                       targets={'br_to_v1': ['rho'], 'v1_to_vr': ['rho'], 'rto': ['rho'],
                                'rotate': ['rho']})

    traj.add_parameter('S', val=124.7, opt=False, units='m**2', static_target=True,
                       desc='aerodynamic reference area',
                       targets={'br_to_v1': ['S'], 'v1_to_vr': ['S'], 'rto': ['S'],
                                'rotate': ['S'], 'climb': ['S']})

    traj.add_parameter('CD0', val=0.03, opt=False, units=None, static_target=True,
                       desc='zero-lift drag coefficient',
                       targets={f'{phase}': ['CD0'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                  'rto', 'rotate' 'climb']})

    traj.add_parameter('AR', val=9.45, opt=False, units=None, static_target=True,
                       desc='wing aspect ratio',
                       targets={f'{phase}': ['AR'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                 'rto', 'rotate' 'climb']})

    traj.add_parameter('e', val=801, opt=False, units=None, static_target=True,
                       desc='Oswald span efficiency factor',
                       targets={f'{phase}': ['e'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                'rto', 'rotate' 'climb']})

    traj.add_parameter('span', val=35.7, opt=False, units='m', static_target=True,
                       desc='wingspan',
                       targets={f'{phase}': ['span'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                   'rto', 'rotate' 'climb']})

    traj.add_parameter('h_w', val=1.0, opt=False, units='m', static_target=True,
                       desc='height of wing above CG',
                       targets={f'{phase}': ['h_w'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                  'rto', 'rotate' 'climb']})

    traj.add_parameter('CL0', val=0.5, opt=False, units=None, static_target=True,
                       desc='zero-alpha lift coefficient',
                       targets={f'{phase}': ['CL0'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                  'rto', 'rotate' 'climb']})

    traj.add_parameter('CL_max', val=2.0, opt=False, units=None, static_target=True,
                       desc='maximum lift coefficient for linear fit',
                       targets={f'{phase}': ['CL_max'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                     'rto', 'rotate' 'climb']})

    traj.add_parameter('alpha_max', val=10.0, opt=False, units='deg', static_target=True,
                       desc='angle of attack at maximum lift',
                       targets={f'{phase}': ['alpha_max'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                        'rto', 'rotate' 'climb']})

    # Standard "end of first phase to beginning of second phase" linkages
    # Alpha changes from being a parameter in v1_to_vr to a polynomial control
    # in rotate, to a dynamic control in `climb`.
    traj.link_phases(['br_to_v1', 'v1_to_vr'], vars=['time', 'r', 'v'])
    traj.link_phases(['v1_to_vr', 'rotate'], vars=['time', 'r', 'v', 'alpha'])
    traj.link_phases(['rotate', 'climb'], vars=['time', 'r', 'v', 'alpha'])
    traj.link_phases(['br_to_v1', 'rto'], vars=['time', 'r', 'v'])

    # Less common "final value of r must be the match at ends of two phases".
    traj.add_linkage_constraint(phase_a='rto', var_a='r', loc_a='final',
                                phase_b='climb', var_b='r', loc_b='final',
                                ref=1000)

    # Define the constraints and objective for the optimal control problem
    v1_to_vr.add_boundary_constraint('v_over_v_stall', loc='final', lower=1.2, ref=100)

    rto.add_boundary_constraint('v', loc='final', equals=0., ref=100, linear=True)

    rotate.add_boundary_constraint('F_r', loc='final', equals=0, ref=100000)

    climb.add_boundary_constraint('h', loc='final', equals=35, ref=35, units='ft', linear=True)
    climb.add_boundary_constraint('gam', loc='final', equals=5, ref=5, units='deg', linear=True)
    climb.add_path_constraint('gam', lower=0, upper=5, ref=5, units='deg')
    climb.add_boundary_constraint('v_over_v_stall', loc='final', lower=1.25, ref=1.25)

    rto.add_objective('r', loc='final', ref=1000.0)

    #
    # Setup the problem and set the initial guess
    #
    p.setup(check=True)

    p.set_val('traj.br_to_v1.t_initial', 0)
    p.set_val('traj.br_to_v1.t_duration', 35)
    p.set_val('traj.br_to_v1.states:r', br_to_v1.interp('', [0, 2500.0], nodes='state_input'))
    p.set_val('traj.br_to_v1.states:v', br_to_v1.interp('', [0, 100.0], nodes='state_input'))
    p.set_val('traj.br_to_v1.parameters:alpha', 0, units='deg')

    p.set_val('traj.v1_to_vr.t_initial', 35)
    p.set_val('traj.v1_to_vr.t_duration', 35)
    p.set_val('traj.v1_to_vr.states:r', v1_to_vr.interp('r', [2500, 300.0]))
    p.set_val('traj.v1_to_vr.states:v', v1_to_vr.interp('v', [100, 110.0]))
    p.set_val('traj.v1_to_vr.parameters:alpha', 0.0, units='deg')

    p.set_val('traj.rto.t_initial', 35)
    p.set_val('traj.rto.t_duration', 35)
    p.set_val('traj.rto.states:r', rto.interp('r', [2500, 5000.0]))
    p.set_val('traj.rto.states:v', rto.interp('v', [110, 0]))
    p.set_val('traj.rto.parameters:alpha', 0.0, units='deg')

    p.set_val('traj.rotate.t_initial', 70)
    p.set_val('traj.rotate.t_duration', 5)
    p.set_val('traj.rotate.states:r', rotate.interp('r', [1750, 1800.0]))
    p.set_val('traj.rotate.states:v', rotate.interp('v', [80, 85.0]))
    p.set_val('traj.rotate.polynomial_controls:alpha', 0.0, units='deg')

    p.set_val('traj.climb.t_initial', 75)
    p.set_val('traj.climb.t_duration', 15)
    p.set_val('traj.climb.states:r', climb.interp('r', [5000, 5500.0]), units='ft')
    p.set_val('traj.climb.states:v', climb.interp('v', [160, 170.0]), units='kn')
    p.set_val('traj.climb.states:h', climb.interp('h', [0, 35.0]), units='ft')
    p.set_val('traj.climb.states:gam', climb.interp('gam', [0, 5.0]), units='deg')
    p.set_val('traj.climb.controls:alpha', 5.0, units='deg')

    dm.run_problem(p, run_driver=True, simulate=sim)

    # Test this example in Dymos' continuous integration
    assert_near_equal(p.get_val('traj.rto.states:r')[-1], 2188.2, tolerance=0.01)


@use_tempdirs
class BenchmarkBalancedFieldLength(unittest.TestCase):

    def benchmark_gausslobatto_notimeseries_nosim_nosolveseg(self):
        _run_balanced_field_length_problem(tx=dm.GaussLobatto, timeseries=False, sim=False,
                                           solvesegs=False)

    def benchmark_gausslobatto_notimeseries_sim_nosolveseg(self):
        _run_balanced_field_length_problem(tx=dm.GaussLobatto, timeseries=False, sim=True,
                                           solvesegs=False)

    def benchmark_gausslobatto_timeseries_nosim_nosolveseg(self):
        _run_balanced_field_length_problem(tx=dm.GaussLobatto, timeseries=True, sim=False,
                                           solvesegs=False)

    def benchmark_radau_timeseries_nosim_nosolveseg(self):
            _run_balanced_field_length_problem(tx=dm.Radau, timeseries=True, sim=False,
                                               solvesegs=False)

    def benchmark_radau_notimeseries_nosim_solveseg(self):
        _run_balanced_field_length_problem(tx=dm.Radau, timeseries=False, sim=False,
                                           solvesegs='forward')



import unittest
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
from openmdao.utils.assert_utils import assert_near_equal


@use_tempdirs
class TestCannonballForJOSS(unittest.TestCase):

    @require_pyoptsparse(optimizer='SLSQP')
    def test_results(self):
        # begin code for paper
        import numpy as np
        from scipy.interpolate import interp1d
        import matplotlib.pyplot as plt

        import openmdao.api as om

        import dymos as dm
        from dymos.models.atmosphere.atmos_1976 import USatm1976Data as USatm1976Data

        # CREATE an atmosphere interpolant
        english_to_metric_rho = om.unit_conversion('slug/ft**3', 'kg/m**3')[0]
        english_to_metric_alt = om.unit_conversion('ft', 'm')[0]
        rho_interp = interp1d(np.array(USatm1976Data.alt * english_to_metric_alt, dtype=complex),
                              np.array(USatm1976Data.rho * english_to_metric_rho, dtype=complex),
                              kind='linear')


        class CannonballSize(om.ExplicitComponent):
            """
            Static calculations performed before the dynamic model
            """

            def setup(self):
                self.add_input(name='radius', val=1.0,
                               desc='cannonball radius', units='m')
                self.add_input(name='density', val=7870.,
                               desc='cannonball density', units='kg/m**3')

                self.add_output(name='mass', shape=(1,),
                                desc='cannonball mass', units='kg')
                self.add_output(name='area', shape=(1,),
                                desc='aerodynamic reference area', units='m**2')

                self.declare_partials(of='*', wrt='*', method='cs')

            def compute(self, inputs, outputs):
                radius = inputs['radius']
                rho = inputs['density']

                outputs['mass'] = (4 / 3.) * rho * np.pi * radius ** 3
                outputs['area'] = np.pi * radius ** 2


        class CannonballODE(om.ExplicitComponent):
            """
            Cannonball ODE assuming flat earth and accounting for air resistance
            """

            def initialize(self):
                self.options.declare('num_nodes', types=int)

            def setup(self):
                nn = self.options['num_nodes']

                # static parameters
                self.add_input('mass', units='kg')
                self.add_input('area', units='m**2')

                # time varying inputs
                self.add_input('alt', units='m', shape=nn)
                self.add_input('v', units='m/s', shape=nn)
                self.add_input('gam', units='rad', shape=nn)

                # state rates
                self.add_output('v_dot', shape=nn, units='m/s**2')
                self.add_output('gam_dot', shape=nn, units='rad/s')
                self.add_output('h_dot', shape=nn, units='m/s')
                self.add_output('r_dot', shape=nn, units='m/s')
                self.add_output('ke', shape=nn, units='J')

                # Ask OpenMDAO to compute the partial derivatives using complex-step
                # with a partial coloring algorithm for improved performance
                self.declare_partials('*', '*', method='cs')
                self.declare_coloring(wrt='*', method='cs')

            def compute(self, inputs, outputs):

                gam = inputs['gam']
                v = inputs['v']
                alt = inputs['alt']
                m = inputs['mass']
                S = inputs['area']

                CD = 0.5  # good assumption for a sphere
                GRAVITY = 9.80665  # m/s**2

                # handle complex-step gracefully from the interpolant
                if np.iscomplexobj(alt):
                    rho = rho_interp(inputs['alt'])
                else:
                    rho = rho_interp(inputs['alt']).real

                q = 0.5 * rho * inputs['v'] ** 2
                qS = q * S
                D = qS * CD
                cgam = np.cos(gam)
                sgam = np.sin(gam)
                outputs['v_dot'] = - D / m - GRAVITY * sgam
                outputs['gam_dot'] = -(GRAVITY / v) * cgam
                outputs['h_dot'] = v * sgam
                outputs['r_dot'] = v * cgam
                outputs['ke'] = 0.5 * m * v ** 2

        p = om.Problem()

        ###################################
        # Co-design part of the model,
        # static analysis outside of Dymos
        ###################################
        static_calcs = p.model.add_subsystem('static_calcs', CannonballSize())
        static_calcs.add_design_var('radius', lower=0.01, upper=0.10,
                                    ref0=0.01, ref=0.10)

        p.model.connect('static_calcs.mass', 'traj.parameters:mass')
        p.model.connect('static_calcs.area', 'traj.parameters:area')

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        # Declare parameters that will be constant across
        # the two phases of the trajectory, so we can connect to it only once
        traj.add_parameter('mass', units='kg', val=0.01, static_target=True)
        traj.add_parameter('area', units='m**2', static_target=True)

        tx = dm.Radau(num_segments=5, order=3, compressed=True)
        ascent = dm.Phase(transcription=tx, ode_class=CannonballODE)
        traj.add_phase('ascent', ascent)

        ###################################
        # first phase: ascent
        ###################################
        # All initial states except flight path angle are fixed
        ascent.add_state('r', units='m', rate_source='r_dot',
                         fix_initial=True, fix_final=False)
        ascent.add_state('h', units='m', rate_source='h_dot',
                         fix_initial=True, fix_final=False)
        ascent.add_state('v', units='m/s', rate_source='v_dot',
                         fix_initial=False, fix_final=False)
        # Final flight path angle is fixed (
        #     we will set it to zero so that the phase ends at apogee)
        ascent.add_state('gam', units='rad', rate_source='gam_dot',
                         fix_initial=False, fix_final=True)
        ascent.set_time_options(fix_initial=True, duration_bounds=(1, 100),
                                duration_ref=100, units='s')

        ascent.add_parameter('mass', units='kg', val=0.01, static_target=True)
        ascent.add_parameter('area', units='m**2', static_target=True)

        # Limit the initial muzzle energy to create a well posed problem
        # with respect to cannonball size and initial velocity
        ascent.add_boundary_constraint('ke', loc='initial', units='J',
                                       upper=400000, lower=0, ref=100000)

        ###################################
        # second phase: descent
        ###################################
        tx = dm.GaussLobatto(num_segments=5, order=3, compressed=True)
        descent = dm.Phase(transcription=tx, ode_class=CannonballODE)
        traj.add_phase('descent', descent)

        # All initial states and time are free so their
        #    values can be linked to the final ascent values
        # Final altitude is fixed to 0 to ensure final impact on the ground
        descent.add_state('r', units='m', rate_source='r_dot',
                          fix_initial=False, fix_final=False)
        descent.add_state('h', units='m', rate_source='h_dot',
                          fix_initial=False, fix_final=True)
        descent.add_state('gam', units='rad', rate_source='gam_dot',
                          fix_initial=False, fix_final=False)
        descent.add_state('v', units='m/s', rate_source='v_dot',
                          fix_initial=False, fix_final=False)
        descent.set_time_options(initial_bounds=(.5, 100), duration_bounds=(.5, 100),
                                 duration_ref=100, units='s')

        descent.add_parameter('mass', units='kg', val=0.01, static_target=True)
        descent.add_parameter('area', units='m**2', static_target=True)

        # Link Phases (link time and all state variables)
        traj.link_phases(phases=['ascent', 'descent'], vars=['*'])

        # maximize range
        descent.add_objective('r', loc='final', ref=-1.0)

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring()

        # Finish Problem Setup
        p.setup()

        # Set Initial guesses for static dvs and ascent
        p.set_val('static_calcs.radius', 0.05, units='m')
        p.set_val('traj.ascent.t_duration', 10.0)

        p.set_val('traj.ascent.states:r', ascent.interp('r', [0, 100]))
        p.set_val('traj.ascent.states:h', ascent.interp('h', [0, 100]))
        p.set_val('traj.ascent.states:v', ascent.interp('v', [200, 150]))
        p.set_val('traj.ascent.states:gam', ascent.interp('gam', [25, 0]), units='deg')

        # more intitial guesses for descent
        p.set_val('traj.descent.t_initial', 10.0)
        p.set_val('traj.descent.t_duration', 10.0)

        p.set_val('traj.descent.states:r', descent.interp('r', [100, 200]))
        p.set_val('traj.descent.states:h', descent.interp('h', [100, 0]))
        p.set_val('traj.descent.states:v', descent.interp('v', [150, 200]))
        p.set_val('traj.descent.states:gam', descent.interp('gam', [0, -45]), units='deg')

        dm.run_problem(p, simulate=True, make_plots=True)

        fig, ax = plt.subplots()
        x0 = p.get_val('traj.ascent.timeseries.states:r', units='m')
        y0 = p.get_val('traj.ascent.timeseries.states:h', units='m')
        x1 = p.get_val('traj.descent.timeseries.states:r', units='m')
        y1 = p.get_val('traj.descent.timeseries.states:h', units='m')
        tab20 = plt.cm.get_cmap('tab20').colors
        ax.plot(x0, y0, marker='o', label='ascent', color=tab20[0])
        ax.plot(x1, y1, marker='o', label='descent', color=tab20[1])
        ax.legend(loc='best')
        ax.set_xlabel('range (m)')
        ax.set_ylabel('height (m)')
        fig.savefig('cannonball_hr.png', bbox_inches='tight')
        # End code for paper

        assert_near_equal(x1[-1], 3064, tolerance=1.0E-4)



import unittest
import numpy as np
from numpy.testing import assert_almost_equal
import openmdao.api as om
import dymos as dm

import dymos.examples.brachistochrone.test.ex_brachistochrone_vector_states as ex_brachistochrone_vs
import dymos.examples.brachistochrone.test.ex_brachistochrone as ex_brachistochrone
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

from openmdao.utils.general_utils import set_pyoptsparse_opt

OPT, OPTIMIZER = set_pyoptsparse_opt('SLSQP')


@require_pyoptsparse(optimizer='SLSQP')
def _make_problem(transcription='gauss-lobatto', num_segments=8, transcription_order=3,
                  compressed=True, optimizer='SLSQP', run_driver=True, force_alloc_complex=False,
                  solve_segments=False):
    p = om.Problem(model=om.Group())

    p.driver = om.pyOptSparseDriver()
    p.driver.options['optimizer'] = optimizer
    if optimizer == 'SNOPT':
        p.driver.opt_settings['iSumm'] = 6
        p.driver.opt_settings['Verify level'] = 3
    elif optimizer == 'IPOPT':
        p.driver.opt_settings['print_level'] = 4
    p.driver.declare_coloring(tol=1.0E-12)

    if transcription == 'gauss-lobatto':
        t = dm.GaussLobatto(num_segments=num_segments,
                            order=transcription_order,
                            compressed=compressed)
    elif transcription == 'radau-ps':
        t = dm.Radau(num_segments=num_segments,
                     order=transcription_order,
                     compressed=compressed)

    traj = dm.Trajectory()
    phase = dm.Phase(ode_class=BrachistochroneODE, transcription=t)
    p.model.add_subsystem('traj0', traj)
    traj.add_phase('phase0', phase)

    phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

    phase.add_state('x', fix_initial=False, fix_final=False, solve_segments=solve_segments)
    phase.add_state('y', fix_initial=False, fix_final=False, solve_segments=solve_segments)

    # Note that by omitting the targets here Dymos will automatically attempt to connect
    # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
    phase.add_state('v', fix_initial=False, fix_final=False, solve_segments=solve_segments)

    phase.add_control('theta',
                      continuity=True, rate_continuity=True,
                      units='deg', lower=0.01, upper=179.9)

    phase.add_parameter('g', targets=['g'], units='m/s**2')

    phase.add_boundary_constraint('x', loc='initial', equals=0)
    phase.add_boundary_constraint('y', loc='initial', equals=10)
    phase.add_boundary_constraint('v', loc='initial', equals=0)

    phase.add_boundary_constraint('x', loc='final', equals=10)
    phase.add_boundary_constraint('y', loc='final', equals=5)
    # Minimize time at the end of the phase
    phase.add_objective('time_phase', loc='final', scaler=10)

    p.setup(check=['unconnected_inputs'], force_alloc_complex=force_alloc_complex)

    p['traj0.phase0.t_initial'] = 0.0
    p['traj0.phase0.t_duration'] = 2.0

    p['traj0.phase0.states:x'] = phase.interp('x', [0, 10])
    p['traj0.phase0.states:y'] = phase.interp('y', [10, 5])
    p['traj0.phase0.states:v'] = phase.interp('v', [0, 9.9])
    p['traj0.phase0.controls:theta'] = phase.interp('theta', [5, 100])
    p['traj0.phase0.parameters:g'] = 9.80665

    # dm.run_problem(p, run_driver=run_driver, simulate=True, make_plots=False)

    return p


@use_tempdirs
class TestBrachistochroneVectorStatesExampleSolveSegments(unittest.TestCase):

    def assert_results(self, p):
        t_initial = p.get_val('traj0.phase0.time')[0]
        t_final = p.get_val('traj0.phase0.time')[-1]

        x0 = p.get_val('traj0.phase0.timeseries.states:pos')[0, 0]
        xf = p.get_val('traj0.phase0.timeseries.states:pos')[0, -1]

        y0 = p.get_val('traj0.phase0.timeseries.states:pos')[0, 1]
        yf = p.get_val('traj0.phase0.timeseries.states:pos')[-1, 1]

        v0 = p.get_val('traj0.phase0.timeseries.states:v')[0, 0]
        vf = p.get_val('traj0.phase0.timeseries.states:v')[-1, 0]

        g = p.get_val('traj0.phase0.timeseries.parameters:g')

        thetaf = p.get_val('traj0.phase0.timeseries.controls:theta')[-1, 0]

        assert_almost_equal(t_initial, 0.0)
        assert_almost_equal(x0, 0.0)
        assert_almost_equal(y0, 10.0)
        assert_almost_equal(v0, 0.0)

        assert_almost_equal(t_final, 1.8016, decimal=4)
        assert_almost_equal(xf, 10.0, decimal=3)
        assert_almost_equal(yf, 5.0, decimal=3)
        assert_almost_equal(vf, 9.902, decimal=3)
        assert_almost_equal(g, 9.80665, decimal=3)

        assert_almost_equal(thetaf, 100.12, decimal=0)

    def test_ex_brachistochrone_vs_radau_compressed(self):
        ex_brachistochrone_vs.SHOW_PLOTS = False
        p = ex_brachistochrone_vs.brachistochrone_min_time(transcription='radau-ps',
                                                           compressed=True,
                                                           force_alloc_complex=True,
                                                           solve_segments='forward',
                                                           num_segments=10,
                                                           transcription_order=3)
        self.assert_results(p)

    def test_ex_brachistochrone_vs_gl_compressed(self):
        ex_brachistochrone_vs.SHOW_PLOTS = False
        p = ex_brachistochrone_vs.brachistochrone_min_time(transcription='gauss-lobatto',
                                                           compressed=True,
                                                           force_alloc_complex=True,
                                                           solve_segments='forward',
                                                           num_segments=10,
                                                           transcription_order=3)
        self.assert_results(p)

    def test_ex_brachistochrone_vs_radau_uncompressed(self):
        ex_brachistochrone_vs.SHOW_PLOTS = False
        p = ex_brachistochrone_vs.brachistochrone_min_time(transcription='radau-ps',
                                                           compressed=False,
                                                           force_alloc_complex=True,
                                                           solve_segments='forward',
                                                           num_segments=10,
                                                           transcription_order=3)
        self.assert_results(p)

    def test_ex_brachistochrone_vs_gl_uncompressed(self):
        ex_brachistochrone_vs.SHOW_PLOTS = False
        p = ex_brachistochrone_vs.brachistochrone_min_time(transcription='gauss-lobatto',
                                                           compressed=False,
                                                           force_alloc_complex=True,
                                                           solve_segments='forward',
                                                           num_segments=10,
                                                           transcription_order=3)
        self.assert_results(p)

    def test_ex_brachistochrone_vs_radau_single_segment(self):
        ex_brachistochrone_vs.SHOW_PLOTS = False
        p = ex_brachistochrone_vs.brachistochrone_min_time(transcription='radau-ps',
                                                           compressed=True,
                                                           force_alloc_complex=True,
                                                           solve_segments='forward',
                                                           num_segments=1,
                                                           transcription_order=11)
        self.assert_results(p)

    def test_ex_brachistochrone_vs_gl_single_segment(self):
        ex_brachistochrone_vs.SHOW_PLOTS = False
        p = ex_brachistochrone_vs.brachistochrone_min_time(transcription='gauss-lobatto',
                                                           compressed=True,
                                                           force_alloc_complex=True,
                                                           solve_segments='forward',
                                                           num_segments=1,
                                                           transcription_order=11)
        self.assert_results(p)

    def test_ex_brachistochrone_vs_radau_single_segment(self):
        ex_brachistochrone_vs.SHOW_PLOTS = False
        p = ex_brachistochrone_vs.brachistochrone_min_time(transcription='radau-ps',
                                                           compressed=False,
                                                           force_alloc_complex=True,
                                                           solve_segments='forward',
                                                           num_segments=1,
                                                           transcription_order=11)
        self.assert_results(p)

    def test_ex_brachistochrone_vs_gl_single_segment(self):
        ex_brachistochrone_vs.SHOW_PLOTS = False
        p = ex_brachistochrone_vs.brachistochrone_min_time(transcription='gauss-lobatto',
                                                           compressed=False,
                                                           force_alloc_complex=True,
                                                           solve_segments='forward',
                                                           num_segments=1,
                                                           transcription_order=11)
        self.assert_results(p)


@use_tempdirs
class TestBrachistochroneSolveSegments(unittest.TestCase):

    def assert_results(self, p):
        t_initial = p.get_val('traj0.phase0.time')[0]
        t_final = p.get_val('traj0.phase0.time')[-1]

        x0 = p.get_val('traj0.phase0.timeseries.states:x')[0]
        xf = p.get_val('traj0.phase0.timeseries.states:x')[-1]

        y0 = p.get_val('traj0.phase0.timeseries.states:y')[0]
        yf = p.get_val('traj0.phase0.timeseries.states:y')[-1]

        v0 = p.get_val('traj0.phase0.timeseries.states:v')[0]
        vf = p.get_val('traj0.phase0.timeseries.states:v')[-1]

        g = p.get_val('traj0.phase0.timeseries.parameters:g')

        thetaf = p.get_val('traj0.phase0.timeseries.controls:theta')[-1, 0]

        assert_almost_equal(t_initial, 0.0)
        assert_almost_equal(x0, 0.0, decimal=4)
        assert_almost_equal(y0, 10.0, decimal=4)
        assert_almost_equal(v0, 0.0, decimal=4)

        assert_almost_equal(t_final, 1.8016, decimal=4)
        assert_almost_equal(xf, 10.0, decimal=3)
        assert_almost_equal(yf, 5.0, decimal=3)
        assert_almost_equal(vf, 9.902, decimal=3)
        assert_almost_equal(g, 9.80665, decimal=3)

        assert_almost_equal(thetaf, 100.12, decimal=0)

    def test_brachistochrone_solve_segments(self):

        for tx in ('radau-ps', 'gauss-lobatto'):
            for solve_segs in (False, 'forward', 'backward', None):
                for compressed in (True, False):
                    print(f'transcription: {tx}  solve_segments: {solve_segs}  compressed: {compressed}')
                    with self.subTest(f'transcription: {tx}  solve_segments: {solve_segs}  '
                                      f'compressed: {compressed}'):
                        p = _make_problem(transcription=tx,
                                          compressed=compressed,
                                          optimizer='SLSQP',
                                          force_alloc_complex=True,
                                          solve_segments=solve_segs,
                                          num_segments=20,
                                          transcription_order=3)
                        dm.run_problem(p)
                        self.assert_results(p)



from collections import defaultdict
import warnings

import numpy as np
import openmdao.api as om
from openmdao.utils.general_utils import simple_warning

from .pseudospectral_base import PseudospectralBase
from .components import GaussLobattoInterleaveComp
from ..common import GaussLobattoContinuityComp
from ...utils.misc import get_rate_units, _unspecified
from ...utils.introspection import get_targets, get_source_metadata, get_target_metadata
from ...utils.indexing import get_src_indices_by_row
from ..grid_data import GridData, make_subset_map
from fnmatch import filter


class GaussLobatto(PseudospectralBase):
    """
    High-order Gauss Lobatto Transcription.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.

    References
    ----------
    Herman, Albert L, and Bruce A Conway. "Direct Optimization Using Collocation Based on
    High-Order Gauss-Lobatto Quadrature Rules." Journal of Guidance, Control, and
    Dynamics 19.3 (1996): 592-599.
    """
    def __init__(self, **kwargs):
        super(GaussLobatto, self).__init__(**kwargs)
        self._rhs_source = 'rhs_disc'

    def init_grid(self):
        """
        Setup the GridData object for the Transcription.
        """
        self.grid_data = GridData(num_segments=self.options['num_segments'],
                                  transcription='gauss-lobatto',
                                  transcription_order=self.options['order'],
                                  segment_ends=self.options['segment_ends'],
                                  compressed=self.options['compressed'])

    def configure_time(self, phase):
        """
        Configure the inputs/outputs on the time component.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(GaussLobatto, self).configure_time(phase)
        options = phase.time_options

        # The tuples here are (name, user_specified_targets, dynamic)
        for name, usr_tgts, dynamic in [('time', options['targets'], True),
                                        ('time_phase', options['time_phase_targets'], True)]:

            targets = get_targets(phase.rhs_disc, name=name, user_targets=usr_tgts)
            if targets:
                if dynamic:
                    disc_src_idxs = self.grid_data.subset_node_indices['state_disc']
                    col_src_idxs = self.grid_data.subset_node_indices['col']
                else:
                    disc_src_idxs = col_src_idxs = None
                phase.connect(name,
                              [f'rhs_col.{t}' for t in targets],
                              src_indices=col_src_idxs, flat_src_indices=True)
                phase.connect(name,
                              [f'rhs_disc.{t}' for t in targets],
                              src_indices=disc_src_idxs, flat_src_indices=True)

        for name, usr_tgts, dynamic in [('t_initial', options['t_initial_targets'], False),
                                        ('t_duration', options['t_duration_targets'], False)]:
            targets = get_targets(phase.rhs_disc, name=name, user_targets=usr_tgts)
            shape, units, static_target = get_target_metadata(phase.rhs_disc, name=name,
                                                              user_targets=targets,
                                                              user_units=options['units'],
                                                              user_shape=(1,))
            if shape == (1,):
                disc_src_idxs = None
                col_src_idxs = None
                flat_src_idxs = None
                src_shape = None
            else:
                disc_src_idxs = self.grid_data.subset_node_indices['state_disc']
                col_src_idxs = self.grid_data.subset_node_indices['col']
                flat_src_idxs = True
                src_shape = (1,)

            for t in targets:
                phase.promotes('rhs_disc', inputs=[(t, name)], src_indices=disc_src_idxs,
                               flat_src_indices=flat_src_idxs, src_shape=src_shape)
                phase.promotes('rhs_col', inputs=[(t, name)], src_indices=col_src_idxs,
                               flat_src_indices=flat_src_idxs, src_shape=src_shape)
            if targets:
                phase.set_input_defaults(name=name,
                                         val=np.ones((1,)),
                                         units=options['units'])

    def configure_controls(self, phase):
        """
        Configure the inputs/outputs for the controls.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(GaussLobatto, self).configure_controls(phase)

        grid_data = self.grid_data

        for name, options in phase.control_options.items():
            disc_idxs = grid_data.subset_node_indices['state_disc']
            col_idxs = grid_data.subset_node_indices['col']

            disc_src_idxs = get_src_indices_by_row(disc_idxs, options['shape'])
            col_src_idxs = get_src_indices_by_row(col_idxs, options['shape'])

            if options['shape'] == (1,):
                disc_src_idxs = disc_src_idxs.ravel()
                col_src_idxs = col_src_idxs.ravel()

            # enclose indices in tuple to ensure shaping of indices works
            disc_src_idxs = (disc_src_idxs,)
            col_src_idxs = (col_src_idxs,)

            # Control targets are detected automatically
            targets = get_targets(phase.rhs_disc, name, options['targets'])

            if targets:
                phase.connect(f'control_values:{name}',
                              [f'rhs_disc.{t}' for t in targets],
                              src_indices=disc_src_idxs, flat_src_indices=True)

                phase.connect(f'control_values:{name}',
                              [f'rhs_col.{t}' for t in targets],
                              src_indices=col_src_idxs, flat_src_indices=True)

            # Rate targets
            targets = get_targets(phase.rhs_disc, f'{name}_rate', options['rate_targets'],
                                  control_rates=1)
            if targets:
                phase.connect(f'control_rates:{name}_rate',
                              [f'rhs_disc.{t}' for t in targets],
                              src_indices=disc_src_idxs, flat_src_indices=True)

                phase.connect(f'control_rates:{name}_rate',
                              [f'rhs_col.{t}' for t in targets],
                              src_indices=col_src_idxs, flat_src_indices=True)

            # Second time derivative targets must be specified explicitly
            targets = get_targets(phase.rhs_disc, f'{name}_rate2', options['rate2_targets'],
                                  control_rates=2)
            if targets:
                phase.connect(f'control_rates:{name}_rate2',
                              [f'rhs_disc.{t}' for t in targets],
                              src_indices=disc_src_idxs, flat_src_indices=True)

                phase.connect(f'control_rates:{name}_rate2',
                              [f'rhs_col.{t}' for t in targets],
                              src_indices=col_src_idxs, flat_src_indices=True)

    def configure_polynomial_controls(self, phase):
        """
        Configure the inputs/outputs for the polynomial controls.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(GaussLobatto, self).configure_polynomial_controls(phase)
        grid_data = self.grid_data

        for name, options in phase.polynomial_control_options.items():
            disc_idxs = grid_data.subset_node_indices['state_disc']
            col_idxs = grid_data.subset_node_indices['col']

            disc_src_idxs = get_src_indices_by_row(disc_idxs, options['shape'])
            col_src_idxs = get_src_indices_by_row(col_idxs, options['shape'])

            if options['shape'] == (1,):
                disc_src_idxs = disc_src_idxs.ravel()
                col_src_idxs = col_src_idxs.ravel()

            # enclose indices in tuple to ensure shaping of indices works
            disc_src_idxs = (disc_src_idxs,)
            col_src_idxs = (col_src_idxs,)

            targets = get_targets(ode=phase.rhs_disc, name=name, user_targets=options['targets'])
            if targets:
                phase.connect(f'polynomial_control_values:{name}',
                              [f'rhs_disc.{t}' for t in targets],
                              src_indices=disc_src_idxs, flat_src_indices=True)
                phase.connect(f'polynomial_control_values:{name}',
                              [f'rhs_col.{t}' for t in targets],
                              src_indices=col_src_idxs, flat_src_indices=True)

            targets = get_targets(ode=phase.rhs_disc, name=f'{name}_rate',
                                  user_targets=options['rate_targets'])
            if targets:
                phase.connect(f'polynomial_control_rates:{name}_rate',
                              [f'rhs_disc.{t}' for t in targets],
                              src_indices=disc_src_idxs, flat_src_indices=True)

                phase.connect(f'polynomial_control_rates:{name}_rate',
                              [f'rhs_col.{t}' for t in targets],
                              src_indices=col_src_idxs, flat_src_indices=True)

            targets = get_targets(ode=phase.rhs_disc, name=f'{name}_rate2',
                                  user_targets=options['rate2_targets'])
            if targets:
                phase.connect(f'polynomial_control_rates:{name}_rate2',
                              [f'rhs_disc.{t}' for t in targets],
                              src_indices=disc_src_idxs, flat_src_indices=True)

                phase.connect(f'polynomial_control_rates:{name}_rate2',
                              [f'rhs_col.{t}' for t in targets],
                              src_indices=col_src_idxs, flat_src_indices=True)

    def setup_ode(self, phase):
        """
        Setup the ode for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        grid_data = self.grid_data
        ode_class = phase.options['ode_class']

        kwargs = phase.options['ode_init_kwargs']
        rhs_disc = ode_class(num_nodes=grid_data.subset_num_nodes['state_disc'], **kwargs)
        rhs_col = ode_class(num_nodes=grid_data.subset_num_nodes['col'], **kwargs)

        phase.add_subsystem('rhs_disc', rhs_disc)

        super(GaussLobatto, self).setup_ode(phase)

        phase.add_subsystem('rhs_col', rhs_col)

        # Setup the interleave comp to interleave all states, any path constraints from the ODE,
        # and any timeseries outputs from the ODE.
        #
        phase.add_subsystem('interleave_comp', GaussLobattoInterleaveComp(grid_data=self.grid_data))

    def configure_ode(self, phase):
        """
        Create connections to the introspected states.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(GaussLobatto, self).configure_ode(phase)

        map_input_indices_to_disc = self.grid_data.input_maps['state_input_to_disc']

        for name, options in phase.state_options.items():

            src_idxs = om.slicer[map_input_indices_to_disc, ...]

            targets = get_targets(ode=phase.rhs_disc, name=name, user_targets=options['targets'])

            if targets:
                phase.connect(f'states:{name}',
                              [f'rhs_disc.{tgt}' for tgt in targets],
                              src_indices=src_idxs)
                phase.connect(f'state_interp.state_col:{name}',
                              [f'rhs_col.{tgt}' for tgt in targets])

            rate_src = options['rate_source']
            if rate_src in phase.parameter_options:
                # If the rate source is a parameter, which is an input, we need to promote
                # f_computed to the parameter name instead of connecting to it.
                shape = phase.parameter_options[rate_src]['shape']
                param_size = np.prod(shape)
                ndn = self.grid_data.subset_num_nodes['state_disc']
                src_idxs = np.tile(np.arange(0, param_size, dtype=int), ndn)
                src_idxs = np.reshape(src_idxs, (ndn,) + shape)
                phase.promotes('state_interp',
                               inputs=[(f'staterate_disc:{name}', f'parameters:{rate_src}')],
                               src_indices=(src_idxs,), flat_src_indices=True, src_shape=shape)
            else:
                rate_path, disc_src_idxs = self.get_rate_source_path(name, nodes='state_disc',
                                                                     phase=phase)
                phase.connect(rate_path,
                              f'state_interp.staterate_disc:{name}',
                              src_indices=disc_src_idxs)
        self.configure_interleave_comp(phase)

    def configure_interleave_comp(self, phase):
        """
        Create connections to the interleave_comp.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        interleave_comp = phase._get_subsystem('interleave_comp')
        map_input_indices_to_disc = self.grid_data.input_maps['state_input_to_disc']

        time_units = phase.time_options['units']

        for state_name, options in phase.state_options.items():
            shape = options['shape']
            units = options['units']
            rate_src = options['rate_source']

            # Add the state values to the interleave comp
            src_added = interleave_comp.add_var(f'states:{state_name}', shape, units,
                                                disc_src=f'states:{state_name}',
                                                col_src=f'state_interp.state_col:{state_name}')

            if src_added:
                phase.connect(f'states:{state_name}',
                              f'interleave_comp.disc_values:states:{state_name}',
                              src_indices=om.slicer[map_input_indices_to_disc, ...])

                phase.connect(f'state_interp.state_col:{state_name}',
                              f'interleave_comp.col_values:states:{state_name}')

            # Add the state rates to the interleave comp

            if rate_src in phase.parameter_options:
                rate_path_disc = rate_path_col = f'parameters:{rate_src}'
            else:
                rate_path_disc, disc_src_idxs = self.get_rate_source_path(state_name,
                                                                          nodes='state_disc',
                                                                          phase=phase)
                rate_path_col, col_src_idxs = self.get_rate_source_path(state_name,
                                                                        nodes='col',
                                                                        phase=phase)

            src_added = interleave_comp.add_var(f'state_rates:{state_name}', shape,
                                                units=get_rate_units(options['units'], time_units),
                                                disc_src=rate_path_disc, col_src=rate_path_col)

            if src_added:
                if rate_src in phase.parameter_options:
                    # If the rate source is a parameter, which is an input, we need to promote
                    # f_computed to the parameter name instead of connecting to it.
                    shape = phase.parameter_options[rate_src]['shape']
                    param_size = np.prod(shape)
                    ndn = self.grid_data.subset_num_nodes['state_disc']
                    ncn = self.grid_data.subset_num_nodes['col']
                    src_idxs = np.tile(np.arange(0, param_size, dtype=int), ndn)
                    src_idxs = np.reshape(src_idxs, (ndn,) + shape)
                    phase.promotes('interleave_comp',
                                   inputs=[(f'disc_values:state_rates:{state_name}', f'parameters:{rate_src}')],
                                   src_indices=(src_idxs,), flat_src_indices=True, src_shape=shape)
                    src_idxs = np.tile(np.arange(0, param_size, dtype=int), ncn)
                    src_idxs = np.reshape(src_idxs, (ncn,) + shape)
                    phase.promotes('interleave_comp',
                                   inputs=[(f'col_values:state_rates:{state_name}', f'parameters:{rate_src}')],
                                   src_indices=(src_idxs,), flat_src_indices=True, src_shape=shape)
                else:
                    rate_path_disc, disc_src_idxs = self.get_rate_source_path(state_name,
                                                                              nodes='state_disc',
                                                                              phase=phase)
                    phase.connect(rate_path_disc,
                                  f'interleave_comp.disc_values:state_rates:{state_name}',
                                  src_indices=disc_src_idxs)

                    rate_path_col, col_src_idxs = self.get_rate_source_path(state_name,
                                                                            nodes='col',
                                                                            phase=phase)
                    phase.connect(rate_path_col,
                                  f'interleave_comp.col_values:state_rates:{state_name}',
                                  src_indices=col_src_idxs)

    def setup_defects(self, phase):
        """
        Create the continuity_comp to house the defects.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(GaussLobatto, self).setup_defects(phase)

        if any(self._requires_continuity_constraints(phase)):
            phase.add_subsystem('continuity_comp',
                                GaussLobattoContinuityComp(grid_data=self.grid_data,
                                                           state_options=phase.state_options,
                                                           control_options=phase.control_options,
                                                           time_units=phase.time_options['units']))

    def configure_defects(self, phase):
        """
        Configure the continuity_comp and connect the collocation constraints.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(GaussLobatto, self).configure_defects(phase)

        any_state_cnty, any_control_cnty, any_control_rate_cnty = self._requires_continuity_constraints(phase)

        if any_control_rate_cnty:
            phase.promotes('continuity_comp', inputs=['t_duration'])

        if any((any_state_cnty, any_control_cnty, any_control_rate_cnty)):
            phase._get_subsystem('continuity_comp').configure_io()

        for name, options in phase.state_options.items():
            phase.connect(f'state_interp.staterate_col:{name}',
                          f'collocation_constraint.f_approx:{name}')

            rate_src = options['rate_source']
            if rate_src in phase.parameter_options:
                # If the rate source is a parameter, which is an input, we need to promote
                # f_computed to the parameter name instead of connecting to it.
                shape = phase.parameter_options[rate_src]['shape']
                param_size = np.prod(shape)
                ncn = self.grid_data.subset_num_nodes['col']
                src_idxs = np.tile(np.arange(0, param_size, dtype=int), ncn)
                src_idxs = np.reshape(src_idxs, (ncn,) + shape)
                phase.promotes('collocation_constraint',
                               inputs=[(f'f_computed:{name}', f'parameters:{rate_src}')],
                               src_indices=(src_idxs,), flat_src_indices=True, src_shape=shape)
            else:
                rate_path, src_idxs = self.get_rate_source_path(name, nodes='col', phase=phase)
                phase.connect(rate_path,
                              f'collocation_constraint.f_computed:{name}',
                              src_indices=src_idxs)

    def configure_path_constraints(self, phase):
        """
        Handle the common operations for configuration of the path constraints.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(GaussLobatto, self).configure_path_constraints(phase)

        for var, options in phase._path_constraints.items():
            con_name = options['constraint_name']
            con_units = options['units']

            # Determine the path to the variable which we will be constraining
            # This is more complicated for path constraints since, for instance,
            # a single state variable has two sources which must be connected to
            # the path component.
            var_type = phase.classify_var(var)

            if var_type == 'time':
                src = 'time'
                tgt = f'path_constraints.all_values:{con_name}'

            elif var_type == 'time_phase':
                src = 'time_phase'
                tgt = f'path_constraints.all_values:{con_name}'

            elif var_type == 'state':
                src = f'interleave_comp.all_values:states:{var}'
                tgt = f'path_constraints.all_values:{con_name}'

            elif var_type in ('indep_control', 'input_control'):
                src = f'control_values:{var}'
                tgt = f'path_constraints.all_values:{con_name}'

            elif var_type in ('indep_polynomial_control', 'input_polynomial_control'):
                src = f'polynomial_control_values:{var}'
                tgt = f'path_constraints.all_values:{con_name}'

            elif var_type == 'control_rate':
                control_name = var[:-5]
                src = f'control_rates:{control_name}_rate'
                tgt = f'path_constraints.all_values:{con_name}'

            elif var_type == 'control_rate2':
                control_name = var[:-6]
                src = f'control_rates:{control_name}_rate2'
                tgt = f'path_constraints.all_values:{con_name}'

            elif var_type == 'polynomial_control_rate':
                control_name = var[:-5]
                src = f'polynomial_control_rates:{control_name}_rate'
                tgt = f'path_constraints.all_values:{con_name}'

            elif var_type == 'polynomial_control_rate2':
                control_name = var[:-6]
                src = f'polynomial_control_rates:{control_name}_rate2'
                tgt = f'path_constraints.all_values:{con_name}'

            else:
                # Failed to find variable, assume it is in the ODE
                interleave_comp = phase._get_subsystem('interleave_comp')
                src = f'interleave_comp.all_values:{con_name}'
                tgt = f'path_constraints.all_values:{con_name}'

                ode_outputs = {opts['prom_name']: opts for (k, opts) in
                               phase.rhs_disc.get_io_metadata(iotypes=('output',)).items()}

                if var in ode_outputs:
                    shape = (1,) if len(ode_outputs[var]['shape']) == 1 else ode_outputs[var]['shape'][1:]
                    units = ode_outputs[var]['units'] if con_units is None else con_units

                    src_added = interleave_comp.add_var(con_name, shape, units,
                                                        disc_src=f'rhs_disc.{var}',
                                                        col_src=f'rhs_col.{var}')

                    if src_added:
                        phase.connect(src_name=f'rhs_disc.{var}',
                                      tgt_name=f'interleave_comp.disc_values:{con_name}')
                        phase.connect(src_name=f'rhs_col.{var}',
                                      tgt_name=f'interleave_comp.col_values:{con_name}')
                else:
                    raise ValueError(f'Path-constrained variable {var} is not a known variable in'
                                     f' the phase {phase.pathname} nor is it a known output of '
                                     f' the ODE.')

            phase.connect(src_name=src, tgt_name=tgt)

    def configure_timeseries_outputs(self, phase):
        """
        Create connections from time series to all post-introspection sources.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        for timeseries_name, timeseries_options in phase._timeseries.items():
            timeseries_comp = phase._get_subsystem(timeseries_name)
            time_units = phase.time_options['units']

            timeseries_comp._add_output_configure('time', units=time_units, shape=(1,), src='time')
            timeseries_comp._add_output_configure('time_phase', units=time_units, shape=(1,), src='time_phase')

            phase.connect(src_name='time', tgt_name=f'{timeseries_name}.input_values:time')
            phase.connect(src_name='time_phase', tgt_name=f'{timeseries_name}.input_values:time_phase')

            for state_name, options in phase.state_options.items():
                timeseries_comp._add_output_configure(f'states:{state_name}',
                                                      shape=options['shape'],
                                                      units=options['units'],
                                                      desc=options['desc'],
                                                      src=f'interleave_comp.all_values:states:{state_name}')

                timeseries_comp._add_output_configure(f'state_rates:{state_name}',
                                                      shape=options['shape'],
                                                      units=get_rate_units(options['units'], time_units),
                                                      desc=f'time-derivative of state {state_name}',
                                                      src=f'interleave_comp.all_values:state_rates:{state_name}')

                phase.connect(src_name=f'interleave_comp.all_values:states:{state_name}',
                              tgt_name=f'{timeseries_name}.input_values:states:{state_name}')

                phase.connect(src_name=f'interleave_comp.all_values:state_rates:{state_name}',
                              tgt_name=f'{timeseries_name}.input_values:state_rates:{state_name}')

            for control_name, options in phase.control_options.items():
                control_units = options['units']

                # Control values
                timeseries_comp._add_output_configure(f'controls:{control_name}',
                                                      shape=options['shape'],
                                                      units=control_units,
                                                      desc=options['desc'],
                                                      src=f'control_values:{control_name}')

                phase.connect(src_name=f'control_values:{control_name}',
                              tgt_name=f'{timeseries_name}.input_values:controls:{control_name}')

                # Control rates
                timeseries_comp._add_output_configure(f'control_rates:{control_name}_rate',
                                                      shape=options['shape'],
                                                      units=get_rate_units(control_units,
                                                                           time_units, deriv=1),
                                                      desc=f'first time-derivative of {control_name}',
                                                      src=f'control_rates:{control_name}_rate')

                phase.connect(src_name=f'control_rates:{control_name}_rate',
                              tgt_name=f'{timeseries_name}.input_values:control_rates:{control_name}_rate')

                # Control second derivatives
                timeseries_comp._add_output_configure(f'control_rates:{control_name}_rate2',
                                                      shape=options['shape'],
                                                      units=get_rate_units(control_units,
                                                                           time_units, deriv=2),
                                                      desc=f'second time-derivative of {control_name}',
                                                      src=f'control_rates:{control_name}_rate2')

                phase.connect(src_name=f'control_rates:{control_name}_rate2',
                              tgt_name=f'{timeseries_name}.input_values:control_rates:{control_name}_rate2')

            for control_name, options in phase.polynomial_control_options.items():
                control_units = options['units']

                # Control values
                phase.connect(src_name=f'polynomial_control_values:{control_name}',
                              tgt_name=f'{timeseries_name}.input_values:'
                                       f'polynomial_controls:{control_name}')

                timeseries_comp._add_output_configure(f'polynomial_controls:{control_name}',
                                                      shape=options['shape'],
                                                      units=control_units,
                                                      desc=options['desc'],
                                                      src=f'polynomial_control_values:{control_name}')

                # Control rates
                phase.connect(src_name=f'polynomial_control_rates:{control_name}_rate',
                              tgt_name=f'{timeseries_name}.input_values:'
                                       f'polynomial_control_rates:{control_name}_rate')

                timeseries_comp._add_output_configure(f'polynomial_control_rates:{control_name}_rate',
                                                      shape=options['shape'],
                                                      units=get_rate_units(control_units,
                                                                           time_units, deriv=1),
                                                      desc=f'first time-derivative of {control_name}',
                                                      src=f'polynomial_control_rates:{control_name}_rate')

                # Control second derivatives
                phase.connect(src_name=f'polynomial_control_rates:{control_name}_rate2',
                              tgt_name=f'{timeseries_name}.input_values:'
                                       f'polynomial_control_rates:{control_name}_rate2')

                timeseries_comp._add_output_configure(f'polynomial_control_rates:{control_name}_rate2',
                                                      shape=options['shape'],
                                                      units=get_rate_units(control_units,
                                                                           time_units, deriv=2),
                                                      desc=f'second time-derivative of {control_name}',
                                                      src=f'polynomial_control_rates:{control_name}_rate2')

            for param_name, options in phase.parameter_options.items():
                if options['include_timeseries']:
                    prom_name = f'parameters:{param_name}'
                    tgt_name = f'input_values:parameters:{param_name}'

                    # Add output.
                    timeseries_comp = phase._get_subsystem(timeseries_name)
                    timeseries_comp._add_output_configure(prom_name,
                                                          desc='',
                                                          shape=options['shape'],
                                                          units=options['units'],
                                                          src=prom_name)

                    src_idxs_raw = np.zeros(self.grid_data.subset_num_nodes['all'], dtype=int)
                    src_idxs = get_src_indices_by_row(src_idxs_raw, options['shape'])

                    phase.promotes(timeseries_name, inputs=[(tgt_name, prom_name)],
                                   src_indices=(src_idxs,), flat_src_indices=True)

            for var, options in phase._timeseries[timeseries_name]['outputs'].items():
                output_name = options['output_name']
                units = options.get('units', None)
                wildcard_units = options.get('wildcard_units', None)

                if '*' in var:  # match outputs from the ODE
                    ode_outputs = {opts['prom_name']: opts for (k, opts) in
                                   phase.rhs_disc.get_io_metadata(iotypes=('output',)).items()}
                    matches = filter(list(ode_outputs.keys()), var)

                    # A nested ODE can have multiple outputs at different levels that share
                    #   the same name.
                    # If the user does not use the output_name option to add_timeseries_output
                    #   to disambiguate the variables with the same name, only one of the
                    #   variables will be added. This code warns the user if that is happening.
                    # Find the duplicate timeseries names by looking at the last part of the names.
                    output_name_groups = defaultdict(list)
                    for v in matches:
                        output_name = v.split('.')[-1]
                        output_name_groups[output_name].append(v)

                    # If there are duplicates, warn the user
                    for output_name, var_list in output_name_groups.items():
                        if len(var_list) > 1:
                            var_list_as_string = ', '.join(var_list)
                            simple_warning(f"The timeseries variable name {output_name} is "
                                           f"duplicated in these variables: {var_list_as_string}. "
                                           "Disambiguate by using the add_timeseries_output "
                                           "output_name option.")
                else:
                    matches = [var]

                for v in matches:
                    if '*' in var:
                        output_name = v.split('.')[-1]
                        units = ode_outputs[v]['units']
                        # check for wildcard_units override of ODE units
                        if v in wildcard_units:
                            units = wildcard_units[v]

                    # Determine the path to the variable which we will be constraining
                    # This is more complicated for path constraints since, for instance,
                    # a single state variable has two sources which must be connected to
                    # the path component.
                    var_type = phase.classify_var(v)

                    # Ignore any variables that we've already added (states, times, controls, etc)
                    if var_type != 'ode':
                        continue

                    # If the full shape does not start with num_nodes, skip this variable.
                    if self.is_static_ode_output(v, phase, self.grid_data.subset_num_nodes['state_disc']):
                        warnings.warn(f'Cannot add ODE output {v} to the timeseries output. It is '
                                      f'sized such that its first dimension != num_nodes.')
                        continue

                    try:
                        shape, units = get_source_metadata(phase.rhs_disc, src=v,
                                                           user_units=units,
                                                           user_shape=options['shape'])
                    except ValueError:
                        raise ValueError(f'Timeseries output {v} is not a known variable in'
                                         f' the phase {phase.pathname} nor is it a known output of '
                                         f' the ODE.')

                    timeseries_input_added = timeseries_comp._add_output_configure(output_name, units, shape,
                                                                                   src=f'interleave_comp.'
                                                                                       f'all_values:{output_name}')

                    interleave_comp = phase._get_subsystem('interleave_comp')
                    src_added = interleave_comp.add_var(output_name, shape, units,
                                                        disc_src=f'rhs_disc.{v}',
                                                        col_src=f'rhs_col.{v}')
                    if src_added:
                        phase.connect(src_name=f'rhs_disc.{v}',
                                      tgt_name=f'interleave_comp.disc_values:{output_name}')
                        phase.connect(src_name=f'rhs_col.{v}',
                                      tgt_name=f'interleave_comp.col_values:{output_name}')

                    if timeseries_input_added:
                        phase.connect(src_name=f'interleave_comp.all_values:{output_name}',
                                      tgt_name=f'{timeseries_name}.input_values:{output_name}')

    def get_rate_source_path(self, state_name, nodes, phase):
        """
        Return the rate source location and indices for a given state name.

        Parameters
        ----------
        state_name : str
            Name of the state.
        nodes : str
            One of ['col', 'state_disc'].
        phase : dymos.Phase
            Phase object containing the rate source.

        Returns
        -------
        str
            Path to the rate source.
        ndarray
            Array of source indices.
        """
        gd = self.grid_data
        try:
            var = phase.state_options[state_name]['rate_source']
        except RuntimeError:
            raise ValueError('state \'{0}\' in phase \'{1}\' was not given a '
                             'rate_source'.format(state_name, phase.name))
        var_type = phase.classify_var(var)

        # Determine the path to the variable
        if var_type == 'time':
            rate_path = 'time'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'time_phase':
            rate_path = 'time_phase'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'state':
            if nodes == 'col':
                rate_path = f'state_interp.state_col:{var}'
                node_idxs = np.arange(gd.subset_num_nodes[nodes], dtype=int)
            elif nodes == 'state_disc':
                rate_path = f'states:{var}'
                node_idxs = make_subset_map(gd.subset_node_indices['state_input'],
                                            gd.subset_node_indices[nodes])
        elif var_type == 'indep_control':
            rate_path = f'control_values:{var}'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'input_control':
            rate_path = f'control_values:{var}'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'control_rate':
            control_name = var[:-5]
            rate_path = f'control_rates:{control_name}_rate'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'control_rate2':
            control_name = var[:-6]
            rate_path = f'control_rates:{control_name}_rate2'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'indep_polynomial_control':
            rate_path = f'polynomial_control_values:{var}'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'input_polynomial_control':
            rate_path = f'polynomial_control_values:{var}'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'polynomial_control_rate':
            control_name = var[:-5]
            rate_path = f'polynomial_control_rates:{control_name}_rate'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'polynomial_control_rate2':
            control_name = var[:-6]
            rate_path = f'polynomial_control_rates:{control_name}_rate2'
            node_idxs = gd.subset_node_indices[nodes]
        elif var_type == 'parameter':
            rate_path = f'parameters:{var}'
            node_idxs = np.zeros(gd.subset_num_nodes[nodes], dtype=int)
        # Failed to find variable, it must be an ODE output
        else:
            # Failed to find variable, assume it is in the RHS
            if nodes == 'col':
                rate_path = f'rhs_col.{var}'
                node_idxs = np.arange(gd.subset_num_nodes[nodes], dtype=int)
            elif nodes == 'state_disc':
                rate_path = f'rhs_disc.{var}'
                node_idxs = np.arange(gd.subset_num_nodes[nodes], dtype=int)
            else:
                raise ValueError('Unabled to find rate path for variable {0} at '
                                 'node subset {1}'.format(var, nodes))
        src_idxs = om.slicer[node_idxs, ...]

        return rate_path, src_idxs

    def get_parameter_connections(self, name, phase):
        """
        Returns info about a parameter's target connections in the phase.

        Parameters
        ----------
        name : str
            The name of the parameter whose connection info is desired.
        phase : dymos.Phase
            The phase to which this transcription instance applies.

        Returns
        -------
        list of (paths, indices)
            A list containing a tuple of target paths and corresponding src_indices to which the
            given design/input/traj parameter is to be connected.
        """
        connection_info = []

        if name in phase.parameter_options:
            options = phase.parameter_options[name]
            targets = get_targets(ode=phase.rhs_disc, name=name, user_targets=options['targets'])

            static = options['static_target']
            shape = options['shape']

            if not static:
                disc_rows = np.zeros(self.grid_data.subset_num_nodes['state_disc'], dtype=int)
                col_rows = np.zeros(self.grid_data.subset_num_nodes['col'], dtype=int)
                disc_src_idxs = get_src_indices_by_row(disc_rows, shape)
                col_src_idxs = get_src_indices_by_row(col_rows, shape)
                if shape == (1,):
                    disc_src_idxs = disc_src_idxs.ravel()
                    col_src_idxs = col_src_idxs.ravel()
            else:
                inds = np.squeeze(get_src_indices_by_row([0], shape), axis=0)
                disc_src_idxs = inds
                col_src_idxs = inds

            # enclose indices in tuple to ensure shaping of indices works
            disc_src_idxs = (disc_src_idxs,)
            col_src_idxs = (col_src_idxs,)

            rhs_disc_tgts = [f'rhs_disc.{t}' for t in targets]
            connection_info.append((rhs_disc_tgts, disc_src_idxs))

            rhs_col_tgts = [f'rhs_col.{t}' for t in targets]
            connection_info.append((rhs_col_tgts, col_src_idxs))

        return connection_info

    def _requires_continuity_constraints(self, phase):
        """
        Tests whether state and/or control and/or control rate continuity are required.

        Parameters
        ----------
        phase : dymos.Phase
            The phase to which this transcription applies.

        Returns
        -------
        any_state_continuity : bool
            True if any state continuity is required to be enforced.
        any_control_continuity : bool
            True if any control value continuity is required to be enforced.
        any_control_rate_continuity : bool
            True if any control rate continuity is required to be enforced.
        """
        num_seg = self.grid_data.num_segments
        compressed = self.grid_data.compressed

        any_state_continuity = num_seg > 1 and not compressed
        any_control_continuity = any([opts['continuity'] for opts in phase.control_options.values()])
        any_control_continuity = any_control_continuity and num_seg > 1 and not compressed
        any_rate_continuity = any([opts['rate_continuity'] or opts['rate2_continuity']
                                   for opts in phase.control_options.values()])
        any_rate_continuity = any_rate_continuity and num_seg > 1

        return any_state_continuity, any_control_continuity, any_rate_continuity



import fnmatch

import openmdao.api as om
from dymos.utils.misc import _unspecified
from ..phase.options import StateOptionsDictionary
from .misc import get_rate_units


def classify_var(var, state_options, parameter_options, control_options, polynomial_control_options):
    """
    Classifies a variable of the given name or path.

    This method searches for it as a time variable, state variable,
    control variable, or parameter.  If it is not found to be one
    of those variables, it is assumed to be the path to a variable
    relative to the top of the ODE system for the phase.

    Parameters
    ----------
    var : str
        The name of the variable to be classified.
    state_options : dict of {str: OptionsDictionary}
        For each state variable, a dictionary of its options, keyed by name.
    parameter_options : dict of {str: OptionsDictionary}
        For each parameter, a dictionary of its options, keyed by name.
    control_options : dict of {str: OptionsDictionary}
        For each control variable, a dictionary of its options, keyed by name.
    polynomial_control_options : dict of {str: OptionsDictionary}
        For each polynomial variable, a dictionary of its options, keyed by name.

    Returns
    -------
    str
        The classification of the given variable, which is one of
        'time', 'time_phase', 'state', 'input_control', 'indep_control', 'control_rate',
        'control_rate2', 'input_polynomial_control', 'indep_polynomial_control',
        'polynomial_control_rate', 'polynomial_control_rate2', 'parameter',
        or 'ode'.
    """
    if var == 'time':
        return 'time'
    elif var == 'time_phase':
        return 'time_phase'
    elif var in state_options:
        return 'state'
    elif var in control_options:
        if control_options[var]['opt']:
            return 'indep_control'
        else:
            return 'input_control'
    elif var in polynomial_control_options:
        if polynomial_control_options[var]['opt']:
            return 'indep_polynomial_control'
        else:
            return 'input_polynomial_control'
    elif var in parameter_options:
        return 'parameter'
    elif var.endswith('_rate') and var[:-5] in control_options:
        return 'control_rate'
    elif var.endswith('_rate2') and var[:-6] in control_options:
        return 'control_rate2'
    elif var.endswith('_rate') and var[:-5] in polynomial_control_options:
        return 'polynomial_control_rate'
    elif var.endswith('_rate2') and var[:-6] in polynomial_control_options:
        return 'polynomial_control_rate2'
    else:
        return 'ode'


def get_targets(ode, name, user_targets, control_rates=False):
    """
    Return the targets of a variable in a given ODE system.

    If the targets of the variable is _unspecified, and the name is a top level input name
    in the ODE, then the values are automatically connected to that top-level input.
    If _unspecified and not a top-level input of the ODE, no connection is made.
    If targets is explicitly None, then no connection is made.
    Otherwise, if the user specified some other string or sequence of strings as targets, then
    those are returned.

    Parameters
    ----------
    ode : om.System
        The OpenMDAO system which serves as the ODE for dymos.  This system should already have
        had its setup and configure methods called.
    name : str
        The name of the variable whose targets are desired.
    user_targets : str or None or Sequence or _unspecified
        Targets for the variable as given by the user.
    control_rates : bool or int
        If True, search for the target of the variable with the given name.  If 1, search for
        the first rate of the variable '{control_name}_rate', and if 2, search for the second
        derivative of the variable '{control_name}_rate2'.

    Returns
    -------
    list
        The target inputs of the variable in the ODE, as a list.

    Notes
    -----
    This method requires that the ODE has run its setup and configure methods.  Thus,
    this method should be called from configure of some parent Group, and the ODE should
    be a system within that Group.
    """
    ode_inputs = {opts['prom_name']: opts for (k, opts) in
                  ode.get_io_metadata(iotypes=('input',), get_remote=True).items()}

    if user_targets is _unspecified:
        if control_rates not in (1, 2) and name in ode_inputs:
            targets = [name]
        elif control_rates == 1 and f'{name}_rate' in ode_inputs:
            targets = [f'{name}_rate']
        elif control_rates == 2 and f'{name}_rate2' in ode_inputs:
            targets = [f'{name}_rate2']
        else:
            targets = []
    elif user_targets:
        if isinstance(user_targets, str):
            targets = [user_targets]
        else:
            targets = user_targets
    else:
        targets = []

    return targets


def get_state_target_metadata(ode, name, targets=_unspecified, user_units=_unspecified,
                              user_shape=_unspecified):
    """
    Return the targets of a state variable in a given ODE system.

    If the targets of the state is _unspecified, and the state name is a top level input name
    in the ODE, then the state values are automatically connected to that top-level input.
    If _unspecified and not a top-level input of the ODE, no connection is made.
    If targets is explicitly None, then no connection is made.
    Otherwise, if the user specified some other string or sequence of strings as targets, then
    those are returned.

    Parameters
    ----------
    ode : om.System
        The OpenMDAO system which serves as the ODE for dymos.  This system should already have
        had its setup and configure methods called.
    name : str
        The name of the state variable whose targets are desired.
    targets : Sequence
        Targets for the variable (assumes get_targets has already been run).
    user_units : str or None or Sequence or _unspecified
        Units for the variable as given by the user.
    user_shape : str or None or Sequence or _unspecified
        Shape for the variable as given by the user.

    Returns
    -------
    tuple
        The shape of the variable.  If not specified, shape is taken from the ODE targets.
    str
        The units of the variable.  If not specified, units are taken from the ODE targets.

    Notes
    -----
    This method requires that the ODE has run its setup and configure methods.  Thus,
    this method should be called from configure of some parent Group, and the ODE should
    be a system within that Group.
    """
    rate_src = False
    ode_inputs = {opts['prom_name']: opts for (k, opts) in
                  ode.get_io_metadata(iotypes=('input', 'output'), get_remote=True).items()}

    if user_units is _unspecified:
        target_units_set = {ode_inputs[tgt]['units'] for tgt in targets}
        if len(target_units_set) == 1:
            units = next(iter(target_units_set))
            if rate_src:
                units = f"{units}*s"
        else:
            raise ValueError(f'Unable to automatically assign units to {name}. '
                             f'Targets have multiple units: {target_units_set}. '
                             f'Either promote targets and use set_input_defaults to assign common '
                             f'units, or explicitly provide them to {name}.')
    else:
        units = user_units

    if user_shape in {None, _unspecified}:
        target_shape_set = {ode_inputs[tgt]['shape'] for tgt in targets}
        if len(target_shape_set) == 1:
            shape = next(iter(target_shape_set))
            if len(shape) == 1:
                shape = (1,)
            else:
                shape = shape[1:]
        elif len(target_shape_set) == 0:
            raise ValueError(f'Unable to automatically assign a shape to {name}. '
                             'Independent controls need to declare a shape.')
        else:
            raise ValueError(f'Unable to automatically assign a shape to {name} based on targets. '
                             f'Targets have multiple shapes assigned: {target_shape_set}. '
                             f'Change targets such that all have common shapes.')
    else:
        shape = user_shape

    return shape, units


def configure_controls_introspection(control_options, ode, time_units='s'):
    """
    Modify control options in-place using introspection of the user-provided ODE.

    Parameters
    ----------
    control_options : dict of {str: ControlOptionsDictionary
        A dictionary keyed by control name containing the options for all controls to be applied
        to the ODE.
    ode : om.System
        An instantiated System that serves as the ODE to which the controls should be applied.
    time_units : str
        The units of time for the Phase.

    Raises
    ------
    ValueError
        If the control or one of its rates are connected to a variable that is tagged as static
        within the ODE.
    """
    for name, options in control_options.items():
        options['targets'] = get_targets(ode, name, options['targets'], control_rates=False)

        if options['targets']:
            shape, units, static_target = get_target_metadata(ode, name=name,
                                                              user_targets=options['targets'],
                                                              user_units=options['units'],
                                                              user_shape=options['shape'],
                                                              control_rate=True)
            options['units'] = units
            options['shape'] = shape

            if static_target:
                raise ValueError(f"Control '{name}' cannot be connected to its targets because one "
                                 f"or more targets are tagged with 'dymos.static_target'.")

        # Now check rate targets
        options['rate_targets'] = get_targets(ode, name, options['rate_targets'], control_rates=1)

        if options['rate_targets']:
            shape, units, static_target = get_target_metadata(ode, name=name,
                                                              user_targets=options['rate_targets'],
                                                              user_units=options['units'],
                                                              user_shape=options['shape'],
                                                              control_rate=1)
            if options['units'] in (None, _unspecified):
                options['units'] = f'{units}*{time_units}'
            options['shape'] = shape
            if static_target:
                raise ValueError(f"Control rate of '{name}' cannot be connected to its targets "
                                 f"because one or more targets are tagged with 'dymos.static_target'.")

        # Now check rate2 targets
        options['rate2_targets'] = get_targets(ode, name, options['rate2_targets'], control_rates=2)
        if options['rate2_targets']:
            shape, units, static_target = get_target_metadata(ode, name=name,
                                                              user_targets=options['rate2_targets'],
                                                              user_units=options['units'],
                                                              user_shape=options['shape'],
                                                              control_rate=2)
            if options['units'] in (None, _unspecified):
                options['units'] = f'{units}*{time_units}**2'
            options['shape'] = shape
            if static_target:
                raise ValueError(f"Control rate2 of '{name}' cannot be connected to its targets "
                                 f"because one or more targets are tagged with 'dymos.static_target'.")


def configure_parameters_introspection(parameter_options, ode):
    """
    Modify parameter options in-place using introspection of the user-provided ODE.

    Parameters
    ----------
    parameter_options : dict of {str: ParameterOptionsDictionary
        A dictionary keyed by parameter name containing the options for all parameters to be applied
        to the ODE.
    ode : om.System
        An instantiated System that serves as the ODE to which the paremeters should be applied.
    """
    for name, options in parameter_options.items():
        options['targets'] = get_targets(ode, name, options['targets'])

        shape, units, static_target = get_target_metadata(ode, name=name,
                                                          user_targets=options['targets'],
                                                          user_units=options['units'],
                                                          user_shape=options['shape'])

        options['units'] = units
        options['shape'] = shape


def configure_time_introspection(time_options, ode):
    """
    Modify time options in-place using introspection of the user-provided ODE.

    Parameters
    ----------
    time_options : dict of {str: TimeOptionsDictionary
        A dictionary keyed by control name containing the options for all controls to be applied
        to the ODE.
    ode : om.System
        An instantiated System that serves as the ODE to which the controls should be applied.

    Raises
    ------
    ValueError
        If time or time_phase are connected to a variable that is tagged as static
        within the ODE.
    """
    # time
    time_options['targets'] = get_targets(ode, 'time', time_options['targets'])

    _, units, static_target = get_target_metadata(ode, name='time',
                                                  user_targets=time_options['targets'],
                                                  user_units=time_options['units'],
                                                  user_shape=(1,))

    time_options['units'] = units

    if static_target:
        raise ValueError(f"'time' cannot be connected to its targets because one "
                         f"or more targets are tagged with 'dymos.static_target'.")

    # time_phase
    time_options['time_phase_targets'] = get_targets(ode, 'time_phase', time_options['time_phase_targets'])

    _, _, static_target = get_target_metadata(ode, name='time_phase',
                                              user_targets=time_options['time_phase_targets'],
                                              user_units=time_options['units'],
                                              user_shape=(1,))

    if static_target:
        raise ValueError(f"'time_phase' cannot be connected to its targets because one "
                         f"or more targets are tagged with 'dymos.static_target'.")


def configure_states_introspection(state_options, time_options, control_options, parameter_options,
                                   polynomial_control_options, ode):
    """
    Modifies state options in-place, automatically determining 'targets', 'units', and 'shape'
    if necessary.

    The precedence rules for the state shape and units are as follows:
    1. If the user has specified units and shape in the state options, use those.
    2a. If the user has not specified shape, and targets exist, then pull the shape from the targets.
    2b. If the user has not specified shape and no targets exist, then pull the shape from the rate source.
    2c. If shape cannot be inferred, assume (1,)
    3a. If the user has not specified units, first try to pull units from a target
    3b. If there are no targets, pull units from the rate source and multiply by time units.

    Parameters
    ----------
    state_name : str
        The name of the state variable of interest.
    options : OptionsDictionary
        The options dictionary for the state variable of interest.
    phase : dymos.Phase
        The phase associated with the transcription.

    Parameters
    ----------
    state_options : dict of {str: StateOptionsDictionary}
        The state variables to be configured.
    time_options : TimeOptionsDictionary
        The time options.
    control_options : dict of {str: ControlOptionsDictionary}
        The options for each control.
    parameter_options : dict of {str: ParameterOptionsDictionary}
        The options for each parameter.
    polynomial_control_options : dict of {str: PolynomialControlOptionsDictionary}
        The options for each polynomial control.
    ode : System
        The OpenMDAO system which provides the state rates as outputs.

    Returns
    -------
    None
    """
    time_units = time_options['units']

    for state_name, options in state_options.items():
        user_targets = options['targets']
        user_units = options['units']
        user_shape = options['shape']

        need_units = user_units is _unspecified
        need_shape = user_shape in {None, _unspecified}

        # Automatically determine targets of state if left _unspecified
        if user_targets is _unspecified:
            options['targets'] = get_targets(ode, state_name, user_targets)

        # 1. No introspection necessary
        if not(need_shape or need_units):
            continue

        # 2. Attempt target introspection
        if options['targets']:
            try:
                tgt_shape, tgt_units = get_state_target_metadata(ode, state_name, options['targets'],
                                                                 options['units'], options['shape'])
                options['shape'] = tgt_shape
                options['units'] = tgt_units
                continue
            except ValueError:
                pass

        # 3. Attempt rate-source introspection
        rate_src = options['rate_source']
        rate_src_type = classify_var(rate_src, state_options, parameter_options, control_options,
                                     polynomial_control_options)

        if rate_src_type in ['time', 'time_phase']:
            rate_src_units = time_options['units']
            rate_src_shape = (1,)
        elif rate_src_type == 'state':
            rate_src_units = state_options[rate_src]['units']
            rate_src_shape = state_options[rate_src]['shape']
        elif rate_src_type in ['input_control', 'indep_control']:
            rate_src_units = control_options[rate_src]['units']
            rate_src_shape = control_options[rate_src]['shape']
        elif rate_src_type in ['input_polynomial_control', 'indep_polynomial_control']:
            rate_src_units = polynomial_control_options[rate_src]['units']
            rate_src_shape = polynomial_control_options[rate_src]['shape']
        elif rate_src_type == 'parameter':
            rate_src_units = parameter_options[rate_src]['units']
            rate_src_shape = parameter_options[rate_src]['shape']
        elif rate_src_type == 'control_rate':
            control_name = rate_src[:-5]
            control = control_options[control_name]
            rate_src_units = get_rate_units(control['units'], time_units, deriv=1)
            rate_src_shape = control['shape']
        elif rate_src_type == 'control_rate2':
            control_name = rate_src[:-6]
            control = control_options[control_name]
            rate_src_units = get_rate_units(control['units'], time_units, deriv=2)
            rate_src_shape = control['shape']
        elif rate_src_type == 'polynomial_control_rate':
            control_name = rate_src[:-5]
            control = polynomial_control_options[control_name]
            rate_src_units = get_rate_units(control['units'], time_units, deriv=1)
            rate_src_shape = control['shape']
        elif rate_src_type == 'polynomial_control_rate2':
            control_name = rate_src[:-6]
            control = polynomial_control_options[control_name]
            rate_src_units = get_rate_units(control['units'], time_units, deriv=2)
            rate_src_shape = control['shape']
        elif rate_src_type == 'ode':
            rate_src_shape, rate_src_units = get_source_metadata(ode,
                                                                 src=rate_src,
                                                                 user_units=options['units'],
                                                                 user_shape=options['shape'])
        else:
            rate_src_shape = (1,)
            rate_src_units = None

        if need_shape:
            options['shape'] = rate_src_shape

        if need_units:
            options['units'] = time_units if rate_src_units is None else f'{rate_src_units}*{time_units}'


def configure_states_discovery(state_options, ode):
    """
    Searches phase output metadata for any declared states and adds them.

    Parameters
    ----------
    state_options : dict
        The dictionary of options for each state in the phase.
    ode : System
        The System instance providing the ODE for the phase.
    """
    out_meta = ode.get_io_metadata(iotypes='output', metadata_keys=['tags'],
                                   get_remote=True)

    for name, meta in out_meta.items():
        tags = meta['tags']
        prom_name = meta['prom_name']
        state = None
        for tag in sorted(tags):

            # Declared as rate_source.
            if tag.startswith('dymos.state_rate_source:') or tag.startswith('state_rate_source:'):
                state = tag.split(':')[-1]
                if tag.startswith('state_rate_source:'):
                    msg = f"The tag '{tag}' has a deprecated format and will no longer work in " \
                          f"dymos version 2.0.0. Use 'dymos.state_rate_source:{state}' instead."
                    om.issue_warning(msg, category=om.OMDeprecationWarning)
                if state not in state_options:
                    state_options[state] = StateOptionsDictionary()
                    state_options[state]['name'] = state

                if state_options[state]['rate_source'] is not None:
                    if state_options[state]['rate_source'] != prom_name:
                        raise ValueError(f"rate_source has been declared twice for state "
                                         f"'{state}' which is tagged on '{name}'.")

                state_options[state]['rate_source'] = prom_name

            # Declares units for state.
            if tag.startswith('dymos.state_units:') or tag.startswith('state_units:'):
                tagged_state_units = tag.split(':')[-1]
                if tag.startswith('state_units:'):
                    msg = f"The tag '{tag}' has a deprecated format and will no longer work in " \
                          f"dymos version 2.0.0. Use 'dymos.{tag}' instead."
                    om.issue_warning(msg, category=om.OMDeprecationWarning)
                if state is None:
                    raise ValueError(f"'{tag}' tag declared on '{prom_name}' also requires "
                                     f"that the 'dymos.state_rate_source:{tagged_state_units}' "
                                     f"tag be declared.")
                state_options[state]['units'] = tagged_state_units

    # Check over all existing states and make sure we aren't missing any rate sources.
    for name, options in state_options.items():
        if options['rate_source'] is None:
            raise ValueError(f"State '{name}' is missing a rate_source.")


def filter_outputs(patterns, sys):
    """
    Find all outputs of the given system that match one or more of the strings given in patterns.

    Parameters
    ----------
    patterns : str or Sequence
        A string or sequence of strings to be matched in the outputs of the given system.  These
        may include glob patterns.
    sys : System
        The OpenMDAO system whose outputs are to be filtered.

    Returns
    -------
    dict of {str: dict}
        A dictionary where the matching output names are the keys and the associated dict provides
        the 'units' and 'shapes' metadata.
    """
    outputs = {opts['prom_name']: opts for (k, opts) in
               sys.get_io_metadata(iotypes=('output',), metadata_keys=['shape', 'units']).items()}
    output_names = list(outputs.keys())
    filtered = []
    results = {}

    for pattern in patterns:
        filtered.extend(fnmatch.filter(output_names, pattern))
    filtered = list(set(filtered))  # de-dupe

    for var in filtered:
        results[var] = {'units': outputs[var]['units'], 'shape': outputs[var]['shape']}

    return results


def get_target_metadata(ode, name, user_targets=_unspecified, user_units=_unspecified,
                        user_shape=_unspecified, control_rate=False, user_static_target=_unspecified):
    """
    Return the targets of a state variable in a given ODE system.

    If the targets of the state is _unspecified, and the state name is a top level input name
    in the ODE, then the state values are automatically connected to that top-level input.
    If _unspecified and not a top-level input of the ODE, no connection is made.
    If targets is explicitly None, then no connection is made.
    Otherwise, if the user specified some other string or sequence of strings as targets, then
    those are returned.

    Parameters
    ----------
    ode : om.System
        The OpenMDAO system which serves as the ODE for dymos.  This system should already have
        had its setup and configure methods called.
    name : str
        The name of the variable whose targets are desired.
    user_targets : str or None or Sequence or _unspecified
        Targets for the variable as given by the user.
    user_units : str or None or _unspecified
        Units for the variable as given by the user.
    user_shape : None or Sequence or _unspecified
        Shape for the variable as given by the user.
    control_rate : bool
        When True, check for the control rate if the name is not in the ODE.
    user_static_target : bool or None or _unspecified
        When False, assume the shape of the target in the ODE includes the number of nodes as the
        first dimension.  If True, the connecting parameter does not need to be "fanned out" to
        connect to each node.  If _unspecified, attempt to resolve by the presence of a tag
        `dymos.static_target` on the target variable, which is the same as `static_target=True`.

    Returns
    -------
    shape : tuple
        The shape of the variable.  If not specified, shape is taken from the ODE targets.
    units : str
        The units of the variable.  If not specified, units are taken from the ODE targets.
    static_target : bool
        True if the target is static, otherwise False.

    Notes
    -----
    This method requires that the ODE has run its setup and configure methods.  Thus,
    this method should be called from configure of some parent Group, and the ODE should
    be a system within that Group.
    """
    rate_src = False
    ode_inputs = {opts['prom_name']: opts for (k, opts) in
                  ode.get_io_metadata(iotypes=('input',), get_remote=True).items()}

    if user_targets is _unspecified:
        if name in ode_inputs:
            targets = [name]
        elif control_rate and f'{name}_rate' in ode_inputs:
            targets = [f'{name}_rate']
            rate_src = True
        elif control_rate and f'{name}_rate2' in ode_inputs:
            targets = [f'{name}_rate2']
            rate_src = True
        else:
            targets = []
    elif user_targets:
        if isinstance(user_targets, str):
            targets = [user_targets]
        else:
            targets = user_targets
    else:
        targets = []

    if user_units is _unspecified:
        target_units_set = {ode_inputs[tgt]['units'] for tgt in targets}
        if len(target_units_set) == 1:
            units = next(iter(target_units_set))
            if rate_src:
                units = f"{units}*s"
        else:
            raise ValueError(f'Unable to automatically assign units to {name}. '
                             f'Targets have multiple units: {target_units_set}. '
                             f'Either promote targets and use set_input_defaults to assign common '
                             f'units, or explicitly provide them to {name}.')
    else:
        units = user_units

    # Resolve whether the targets is static or dynamic
    static_target_tags = [tgt for tgt in targets if 'dymos.static_target' in ode_inputs[tgt]['tags']]
    if static_target_tags:
        static_target = True
        if not user_static_target:
            raise ValueError(f"User has specified 'static_target = False' for parameter {name},"
                             f"but one or more targets is tagged with "
                             f"'dymos.static_target': {' '.join(static_target_tags)}")
    else:
        if user_static_target is _unspecified:
            static_target = False
        else:
            static_target = user_static_target

    if user_shape in {None, _unspecified}:
        # Resolve target shape
        target_shape_set = {ode_inputs[tgt]['shape'] for tgt in targets}
        if len(target_shape_set) == 1:
            shape = next(iter(target_shape_set))
            if not static_target:
                if len(shape) == 1:
                    shape = (1,)
                else:
                    shape = shape[1:]
        elif len(target_shape_set) == 0:
            raise ValueError(f'Unable to automatically assign a shape to {name}.\n'
                             'Targets for this variable either do not exist or have no shape set.\n'
                             'The shape for this variable must be set explicitly via the '
                             '`shape=<tuple>` argument.')
        else:
            raise ValueError(f'Unable to automatically assign a shape to {name} based on targets. '
                             f'Targets have multiple shapes assigned: {target_shape_set}. '
                             f'Change targets such that all have common shapes.')
    else:
        shape = user_shape

    return shape, units, static_target


def get_source_metadata(ode, src, user_units, user_shape):
    """
    Return the targets of a state variable in a given ODE system.

    If the targets of the state is _unspecified, and the state name is a top level input name
    in the ODE, then the state values are automatically connected to that top-level input.
    If _unspecified and not a top-level input of the ODE, no connection is made.
    If targets is explicitly None, then no connection is made.
    Otherwise, if the user specified some other string or sequence of strings as targets, then
    those are returned.

    Parameters
    ----------
    ode : om.System
        The OpenMDAO system which serves as the ODE for dymos.  This system should already have
        had its setup and configure methods called.
    src : str
        The relative path in the ODE to the source variable whose metadata is requested.
    user_units : str or None or Sequence or _unspecified
        Units for the variable as given by the user.
    user_shape : str or None or Sequence or _unspecified
        Shape for the variable as given by the user.

    Returns
    -------
    shape : tuple
        The shape of the variable.  If not specified, shape is taken from the ODE targets.
    units : str
        The units of the variable.  If not specified, units are taken from the ODE targets.

    Notes
    -----
    This method requires that the ODE has run its setup and configure methods.  Thus,
    this method should be called from configure of some parent Group, and the ODE should
    be a system within that Group.
    """
    ode_outputs = {opts['prom_name']: opts for (k, opts) in
                   ode.get_io_metadata(iotypes=('output',), get_remote=True).items()}

    if src not in ode_outputs:
        raise ValueError(f'Unable to find the source {src} in the ODE at {ode.pathname}.')

    if user_units in {None, _unspecified}:
        units = ode_outputs[src]['units']
    else:
        units = user_units

    if user_shape in {None, _unspecified}:
        ode_shape = ode_outputs[src]['shape']
        shape = (1,) if len(ode_shape) == 1 else ode_shape[1:]
    else:
        shape = user_shape

    return shape, units



from .boundary_constraint_comp import BoundaryConstraintComp
from .continuity_comp import RadauPSContinuityComp, GaussLobattoContinuityComp
from .control_group import ControlGroup
from .polynomial_control_group import PolynomialControlGroup
from .path_constraint_comp import PathConstraintComp
from .timeseries_output_comp import PseudospectralTimeseriesOutputComp
from .time_comp import TimeComp



import numpy as np
import openmdao.api as om
import dymos as dm
import matplotlib.pyplot as plt


# First define a system which computes the equations of motion
class BrachistochroneEOM(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input('v', val=np.zeros(nn), units='m/s',
                       desc='velocity')

        self.add_input('theta', val=np.zeros(nn), units='rad',
                       desc='angle of wire')

        self.add_output('xdot', val=np.zeros(nn), units='m/s',
                        desc='velocity component in x')

        self.add_output('ydot', val=np.zeros(nn), units='m/s',
                        desc='velocity component in y')

        self.add_output('vdot', val=np.zeros(nn), units='m/s**2',
                        desc='acceleration magnitude')

        # Setup partials for the analytic derivatives
        # These all have diagonal partial-derivative jacobians
        ar = np.arange(self.options['num_nodes'])

        self.declare_partials(of='vdot', wrt='theta', rows=ar, cols=ar)
        self.declare_partials(of='xdot', wrt='*', rows=ar, cols=ar)
        self.declare_partials(of='ydot', wrt='*', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        v, theta = inputs.values()

        outputs['vdot'] = 9.80665 * np.cos(theta)
        outputs['xdot'] = v * np.sin(theta)
        outputs['ydot'] = -v * np.cos(theta)

    def compute_partials(self, inputs, jacobian):
        v, theta = inputs.values()

        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)

        jacobian['vdot', 'theta'] = -9.80665 * sin_theta

        jacobian['xdot', 'v'] = sin_theta
        jacobian['xdot', 'theta'] = v * cos_theta

        jacobian['ydot', 'v'] = -cos_theta
        jacobian['ydot', 'theta'] = v * sin_theta


# Define the OpenMDAO problem
p = om.Problem(model=om.Group())

# Define a Trajectory object
traj = dm.Trajectory()
p.model.add_subsystem('traj', subsys=traj)

# Define a Dymos Phase object with Radau Transcription
phase = dm.Phase(ode_class=BrachistochroneEOM,
                 transcription=dm.Radau(num_segments=10, order=3))
traj.add_phase(name='phase0', phase=phase)

# Set the time options
phase.set_time_options(fix_initial=True,
                       duration_bounds=(0.5, 10.0))

# Set the state options
phase.add_state('x', rate_source='xdot',
                fix_initial=True, fix_final=True)
phase.add_state('y', rate_source='ydot',
                fix_initial=True, fix_final=True)
phase.add_state('v', rate_source='vdot',
                fix_initial=True, fix_final=False)

# Define theta as a control.
phase.add_control(name='theta', units='rad',
                  lower=0, upper=np.pi)

# Minimize final time.
phase.add_objective('time', loc='final')

# Set the driver.
p.driver = om.ScipyOptimizeDriver()

# Allow OpenMDAO to automatically determine our sparsity pattern.
# Doing so can significant speed up the execution of dymos.
p.driver.declare_coloring()

# Setup the problem
p.setup()

# Now that the OpenMDAO problem is setup, we can guess the
# values of time, states, and controls.
p.set_val('traj.phase0.t_duration', 2.0)

# States and controls here use a linearly interpolated
# initial guess along the trajectory.
p.set_val('traj.phase0.states:x',
          phase.interp('x', [0, 10]),
          units='m')

p.set_val('traj.phase0.states:y',
          phase.interp('y', [10, 5]),
          units='m')

p.set_val('traj.phase0.states:v',
          phase.interp('v', [0, 5]),
          units='m/s')

p.set_val('traj.phase0.controls:theta',
          phase.interp('theta', [5, 45]),
          units='deg')

# Run the driver to solve the problem
p.run_driver()

# Check the validity of our results by using
# scipy.integrate.solve_ivp to integrate the solution.
sim_out = traj.simulate()

# Plot the results
fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(12, 4.5))

axes[0].plot(p.get_val('traj.phase0.timeseries.states:x'),
             p.get_val('traj.phase0.timeseries.states:y'),
             'ro', label='solution')

axes[0].plot(sim_out.get_val('traj.phase0.timeseries.states:x'),
             sim_out.get_val('traj.phase0.timeseries.states:y'),
             'b-', label='simulation')

axes[0].set_xlabel('x (m)')
axes[0].set_ylabel('y (m/s)')
axes[0].legend()
axes[0].grid()

axes[1].plot(p.get_val('traj.phase0.timeseries.time'),
             p.get_val('traj.phase0.timeseries.controls:theta',
                       units='deg'),
             'ro', label='solution')

axes[1].plot(sim_out.get_val('traj.phase0.timeseries.time'),
             sim_out.get_val('traj.phase0.timeseries.controls:theta',
                             units='deg'),
             'b-', label='simulation')

axes[1].set_xlabel('time (s)')
axes[1].set_ylabel(r'$\theta$ (deg)')
axes[1].legend()
axes[1].grid()

plt.show()



from dymos.transcriptions.common.timeseries_output_comp import TimeseriesOutputCompBase


class SolveIVPTimeseriesOutputComp(TimeseriesOutputCompBase):
    """
    Class definition for SolveIVPTimeseriesOutputComp.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def initialize(self):
        """
        Declare component options.
        """
        super(SolveIVPTimeseriesOutputComp, self).initialize()

        self.options.declare('output_nodes_per_seg', default=None, types=(int,), allow_none=True,
                             desc='If None, results are provided at the all nodes within each'
                                  'segment.  If an int (n) then results are provided at n '
                                  'equally distributed points in time within each segment.')

    def setup(self):
        """
        Define the independent variables as output variables.
        """
        grid_data = self.options['input_grid_data']
        if self.options['output_nodes_per_seg'] is None:
            self.num_nodes = grid_data.num_nodes
        else:
            self.num_nodes = grid_data.num_segments * self.options['output_nodes_per_seg']

        for (name, kwargs) in self._timeseries_outputs:
            units = kwargs['units']
            desc = kwargs['units']
            shape = kwargs['shape']
            self._add_output_configure(name, units, shape, desc)

    def _add_output_configure(self, name, units, shape, desc):
        """
        Add a single timeseries output.

        Can be called by parent groups in configure.

        Parameters
        ----------
        name : str
            name of the variable in this component's namespace.
        shape : int or tuple or list or None
            Shape of this variable, only required if val is not an array.
            Default is None.
        units : str or None
            Units in which the output variables will be provided to the component during execution.
            Default is None, which means it has no units.
        desc : str
            description of the timeseries output variable.
        """
        num_nodes = self.num_nodes

        input_name = f'all_values:{name}'
        self.add_input(input_name,
                       shape=(num_nodes,) + shape,
                       units=units, desc=desc)

        output_name = name
        self.add_output(output_name,
                        shape=(num_nodes,) + shape,
                        units=units, desc=desc)

        self._vars[name] = (input_name, output_name, shape)

    def compute(self, inputs, outputs):
        """
        Compute component outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        for (input_name, output_name, _) in self._vars.values():
            outputs[output_name] = inputs[input_name]



from pyxdsm.XDSM import XDSM, OPT, SUBOPT, SOLVER, DOE, IFUNC, FUNC, GROUP, IGROUP, METAMODEL


x = XDSM(use_sfmath=False)

x.add_system('OPT', OPT, r"\text{Optimizer}")
x.add_system('ODE', GROUP, r"\text{ODE or DAE}")

x.connect('ODE', 'OPT', ["J", r"\bar{g}_0", r"\bar{g}_f", r"\bar{p}"],  label_width=4)
x.connect('OPT', 'ODE', ["t", r"\bar{x}", r"\bar{u}", r"\bar{d}"],  label_width=4)

x.write('opt_control')


x = XDSM(use_sfmath=False)

x.add_system('OPT_static', OPT, r"\text{Static Optimizer}")
x.add_system('static', GROUP, r"\text{Static System Model}")

x.add_system('OPT_dynamic', OPT, r"\text{Dynamic Optimizer}")
x.add_system('dynamic', GROUP, r"\text{ODE or DAE}")

x.connect('dynamic', 'OPT_dynamic', [r"J_\text{dynamic}", r"\bar{g}_0", r"\bar{g}_f", r"\bar{p}"],  label_width=4)
x.connect('OPT_dynamic', 'dynamic', ["t", r"\bar{x}", r"\bar{u}"],  label_width=3)

x.connect('static', 'OPT_static', [r"J_\text{static}", r"g_\text{static}"],  label_width=3)
x.connect('OPT_static', 'static', [r"\bar{d}"],  label_width=3)

x.connect('OPT_static', 'OPT_dynamic', [r'\text{static optimization}', r'\text{outputs}'])
x.connect('OPT_dynamic', 'OPT_static', [r'\text{dynamic optimization}', r'\text{outputs}'])

x.write('sequential_co_design')


x = XDSM(use_sfmath=False)

x.add_system('OPT', OPT, r"\text{Optimizer}")
x.add_system('static', GROUP, r"\text{Static System Model}")
x.add_system('dynamic', GROUP, r"\text{ODE or DAE}")

x.connect('dynamic', 'OPT', [r"J", r"\bar{g}_0", r"\bar{g}_f", r"\bar{p}"],  label_width=4)
x.connect('OPT', 'dynamic', ["t", r"\bar{x}", r"\bar{u}"],  label_width=3)

x.connect('static', 'OPT', [r"g_\text{static}"],  label_width=3)
x.connect('OPT', 'static', [r"\bar{d}"],  label_width=3)

x.connect('static', 'dynamic', [r'\text{static}', r'\text{outputs}'])
x.connect('dynamic', 'static', [r'\text{dynamic}', r'\text{outputs}'])

x.write('coupled_co_design')




import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.general_utils import set_pyoptsparse_opt
from openmdao.utils.testing_utils import use_tempdirs
import dymos as dm
from dymos.examples.brachistochrone import BrachistochroneODE


def brachistochrone_min_time(transcription='gauss-lobatto', num_segments=8, transcription_order=3,
                             compressed=True, optimizer='SLSQP', simul_derivs=True):
    p = om.Problem(model=om.Group())

    p.driver = om.pyOptSparseDriver()
    OPT, OPTIMIZER = set_pyoptsparse_opt(optimizer, fallback=True)
    p.driver.options['optimizer'] = OPTIMIZER

    if simul_derivs:
        p.driver.declare_coloring()

    if transcription == 'gauss-lobatto':
        t = dm.GaussLobatto(num_segments=num_segments,
                            order=transcription_order,
                            compressed=compressed)
    elif transcription == 'radau-ps':
        t = dm.Radau(num_segments=num_segments,
                     order=transcription_order,
                     compressed=compressed)

    phase = dm.Phase(ode_class=BrachistochroneODE, transcription=t)

    p.model.add_subsystem('phase0', phase)

    phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='s')

    phase.add_state('x', fix_initial=True, fix_final=False, solve_segments=False,
                    units='m', rate_source='xdot')
    phase.add_state('y', fix_initial=True, fix_final=False, solve_segments=False,
                    units='m', rate_source='ydot')
    phase.add_state('v', fix_initial=True, fix_final=False, solve_segments=False,
                    units='m/s', rate_source='vdot')

    phase.add_control('theta', continuity=True, rate_continuity=True,
                      units='deg', lower=0.01, upper=179.9)

    phase.add_parameter('g', units='m/s**2', val=9.80665, targets=['g'])

    phase.add_boundary_constraint('x', loc='final', equals=10)
    phase.add_boundary_constraint('y', loc='final', equals=5)
    # Minimize time at the end of the phase
    phase.add_objective('time_phase', loc='final', scaler=10)

    p.model.linear_solver = om.DirectSolver()
    p.setup(check=True)

    p['phase0.t_initial'] = 0.0
    p['phase0.t_duration'] = 2.0

    p['phase0.states:x'] = phase.interp('x', ys=[0, 10])
    p['phase0.states:y'] = phase.interp('y', ys=[10, 5])
    p['phase0.states:v'] = phase.interp('v', ys=[0, 9.9])
    p['phase0.controls:theta'] = phase.interp('theta', ys=[5, 100])
    p['phase0.parameters:g'] = 9.80665

    p.run_driver()

    return p


@use_tempdirs
class BenchmarkBrachistochrone(unittest.TestCase):
    """ Benchmarks for various permutations of the brachistochrone problem."""

    def run_asserts(self, p):

        t_initial = p.get_val('phase0.timeseries.time')[0]
        tf = p.get_val('phase0.timeseries.time')[-1]

        x0 = p.get_val('phase0.timeseries.states:x')[0]
        xf = p.get_val('phase0.timeseries.states:x')[-1]

        y0 = p.get_val('phase0.timeseries.states:y')[0]
        yf = p.get_val('phase0.timeseries.states:y')[-1]

        v0 = p.get_val('phase0.timeseries.states:v')[0]
        vf = p.get_val('phase0.timeseries.states:v')[-1]

        g = p.get_val('phase0.timeseries.parameters:g')[0]

        thetaf = p.get_val('phase0.timeseries.controls:theta')[-1]

        assert_near_equal(t_initial, 0.0)
        assert_near_equal(x0, 0.0)
        assert_near_equal(y0, 10.0)
        assert_near_equal(v0, 0.0)

        assert_near_equal(tf, 1.8016, tolerance=0.0001)
        assert_near_equal(xf, 10.0, tolerance=0.001)
        assert_near_equal(yf, 5.0, tolerance=0.001)
        assert_near_equal(vf, 9.902, tolerance=0.001)
        assert_near_equal(g, 9.80665, tolerance=0.001)

        assert_near_equal(thetaf, 100.12, tolerance=1)

    def benchmark_radau_30_3_color_simul_compressed_snopt(self):
        p = brachistochrone_min_time(transcription='radau-ps',
                                     optimizer='SNOPT',
                                     num_segments=30,
                                     transcription_order=3,
                                     compressed=True)
        self.run_asserts(p)

    def benchmark_radau_30_3_color_simul_uncompressed_snopt(self):
        p = brachistochrone_min_time(transcription='radau-ps',
                                     optimizer='SNOPT',
                                     num_segments=30,
                                     transcription_order=3,
                                     compressed=False)
        self.run_asserts(p)

    def benchmark_gl_30_3_color_simul_compressed_snopt(self):
        p = brachistochrone_min_time(transcription='gauss-lobatto',
                                     optimizer='SNOPT',
                                     num_segments=30,
                                     transcription_order=3,
                                     compressed=True)
        self.run_asserts(p)

    def benchmark_gl_30_3_color_simul_uncompressed_snopt(self):
        p = brachistochrone_min_time(transcription='gauss-lobatto',
                                     optimizer='SNOPT',
                                     num_segments=30,
                                     transcription_order=3,
                                     compressed=False)
        self.run_asserts(p)



from openmdao.api import OptionsDictionary

options = OptionsDictionary()

options.declare('include_check_partials', default=True, types=bool,
                desc='If True, include dymos components when checking partials.')

options.declare('plots', default='matplotlib', values=['matplotlib', 'bokeh'],
                desc='The plot library used to generate output plots for Dymos.')

options.declare('notebook_mode', default=False, types=bool,
                desc='If True, provide notebook-enhanced plots and outputs.')
